# Copyright (c) 2019-2023, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from typing import Tuple

from grid2op.Exceptions import (
    HandlerError
)
from grid2op.Chronics.handlers.baseHandler import BaseHandler
from grid2op.Chronics.handlers.csvHandler import CSVHandler


class CSVForecastHandler(CSVHandler):
    """Reads and produce time series if given by a csv file (possibly compressed).
    
    The separator used can be specified as input. 
    
    The file name should match the "array_name":
    for example if the data you want to use for "load_p_forecasted" in the environment
    are in the file "my_load_p_forecast.csv.bz2" should name this handler 
    "my_load_p_forecast" and not "load_p" nor "my_load_p_forecast.csv" nor
    "my_load_p_forecast.csv.bz2"
    
    The csv should be structured as follow:
    
    - it should not have any "index" or anything, only data used directly
      by grid2op (so only "active loads" if this handler is responsible 
      for the generation of "load_p")
    - Each element (for example a load) is represented by a `column`.
    - It should have a header with the name of the elements it "handles" and 
      this name should match the one in the environment. For example 
      if "load_1_0" is the name of a load and you read data for "load_p"
      or "load_q" then one column of your csv should be named "load_1_0".
    - only floating point numbers should be present in the data (no bool, string
      and integers will be casted to float)
    
    The structuration of the rows are a bit different than for :class:`CSVHandler`
    because this class can read "multiple steps ahead forecast", provided that
    it knows for how many different horizons forecasts are made.
    
    Let's take the example that forecast are available for h = 5, 10 and 15
    minutes ahead (so for the next, next next and next next next steps). In this case:
    
    - the first row (not counting the header) will be the forecast made
      for h = 5 at the first step: the forecasts available at t=0 for t=5mins
    - the second row will be the forecasts made
      for h = 10 at the first step: the forecasts available at t=0 for t=10mins
    - the third row will be the forecasts made
      for h = 15 at the first step: the forecasts available at t=0 for t=15mins
    - the fourth row will be the forecasts made
      for h = 5 at the second step: the forecasts available at t=5 for t=10mins
    - the fifth row will be the forecasts made
      for h = 10 at the second step: the forecasts available at t=5 for t=15mins
    - etc.
      
    .. warning::
        Use this class only for the FORECAST data ("load_p_forecasted", 
        "load_q_forecasted", "prod_p_forecasted" or "prod_v_forecasted") and 
        not for maintenance (in this case
        use :class:`CSVMaintenanceHandler`) nor for 
        environment data (in this case use :class:`CSVHandler`) 
    
    This is the default way to provide data to grid2op and its used for
    most l2rpn environments when forecasts are available.
    
    .. note::
        The current implementation heavily relies on the fact that the 
        :func:`CSVForecastHandler.forecast` method is called
        exactly once per horizon and per step.
    
    """
    def __init__(self,
                 array_name,
                 sep=";",
                 chunk_size=None,
                 max_iter=-1) -> None:
        super().__init__(array_name, sep, chunk_size, max_iter)
        self._h_forecast = None
        self._nb_row_per_step = 1
        
    def load_next(self, dict_):
        raise HandlerError("You should only use this class for FORECAST data, and not for ENVIRONMENT data. "
                           "Please consider using `CSVHandler` (`from grid2op.Chronics.handlers import CSVHandler`) "
                           "for your environment data.")
    
    def set_chunk_size(self, chunk_size):
        super().set_chunk_size(self._nb_row_per_step * int(chunk_size))
        
    def set_max_iter(self, max_iter):
        super().set_max_iter(self._nb_row_per_step * int(max_iter))
    
    def set_h_forecast(self, h_forecast):
        super().set_h_forecast(h_forecast)
        self._nb_row_per_step = len(self._h_forecast) 
        
    def get_available_horizons(self):
        # skip the definition in CSVHandler to jump to the level "above"
        return super(CSVHandler, self).get_available_horizons()
        
    def forecast(self,
                 forecast_horizon_id : int,
                 inj_dict_env : dict,
                 inj_dict_previous_forecast : dict,
                 # eg gen_p_handler if this is set to gen_p_for_handler:
                 env_handler : "BaseHandler",  
                 # list of the 4 env handlers: (load_p_handler, load_q_handler, gen_p_handler, gen_v_handler)
                 env_handlers : Tuple["BaseHandler", "BaseHandler", "BaseHandler", "BaseHandler"]):
        res = super().load_next(inj_dict_previous_forecast)
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Exceptions import PlotError
from grid2op.Plot.PlotPlotly import PlotPlotly
from grid2op.Plot.PlotMatplotlib import PlotMatplotlib
from grid2op.Plot.PlotPyGame import PlotPyGame

from grid2op.Exceptions.PlotExceptions import PyGameQuit


class Plotting:
    """
    INTERNAL

    .. warning:: /!\\\\ This class is deprecated /!\\\\

    """

    allwed_display_mod = {
        "pygame": PlotPyGame,
        "plotly": PlotPlotly,
        "matplotlib": PlotMatplotlib,
    }

    def __init__(
        self,
        observation_space,
        display_mod="plotly",
        substation_layout=None,
        radius_sub=20.0,
        load_prod_dist=70.0,
        bus_radius=6.0,
    ):
        if display_mod not in self.allwed_display_mod:
            raise PlotError(
                'Only avaible plot mod are "{}". You specified "{}" which is not supported.'
                "".format(self.allwed_display_mod, display_mod)
            )

        cls_ = self.allwed_display_mod[display_mod]
        self.displ_backend = cls_(
            observation_space,
            substation_layout=substation_layout,
            radius_sub=radius_sub,
            load_prod_dist=load_prod_dist,
            bus_radius=bus_radius,
        )
        self.display_mod = display_mod

    def _display_fig(self, fig, display):
        if display:
            if self.display_mod == "plotly":
                fig.show()
            elif self.display_mod == "matplotlib":
                fig, ax = fig
                fig.show()

    def plot_layout(
        self, fig=None, reward=None, done=None, timestamp=None, display=True
    ):
        try:
            fig = self.displ_backend.plot_layout(
                fig=fig, reward=reward, done=done, timestamp=timestamp
            )
            self._display_fig(fig, display=display)
        except PyGameQuit:
            pass
        return fig

    def plot_info(
        self,
        fig=None,
        line_info=None,
        load_info=None,
        gen_info=None,
        sub_info=None,
        colormap=None,
        display=True,
    ):
        try:
            fig = self.displ_backend.plot_info(
                fig=fig,
                line_info=line_info,
                load_info=load_info,
                gen_info=gen_info,
                sub_info=sub_info,
                colormap=colormap,
            )
            self._display_fig(fig, display=display)
        except PyGameQuit:
            pass
        return fig

    def plot_obs(
        self,
        observation,
        fig=None,
        reward=None,
        done=None,
        timestamp=None,
        line_info="rho",
        load_info="p",
        gen_info="p",
        colormap="line",
        display=True,
    ):
        try:
            fig = self.displ_backend.plot_obs(
                observation,
                fig=fig,
                reward=reward,
                done=done,
                timestamp=timestamp,
                line_info=line_info,
                load_info=load_info,
                gen_info=gen_info,
                colormap=colormap,
            )
            self._display_fig(fig, display=display)
        except PyGameQuit:
            pass

        return fig



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np
import copy
import warnings

from grid2op.dtypes import dt_float, dt_int
from grid2op.Backend import Backend
from grid2op.Exceptions import Grid2OpException, BackendError

ERROR_NB_ELEMENTS = (
    "Impossible to make a BackendConverter with backends having different number of {}."
)
ERROR_ELEMENT_CONNECTED = (
    "No {0} connected at substation {1} for the target backend while a {0} is connected "
    "at substation {2} for the source backend"
)
ERROR_INVALID_VECTOR = "invalid vector: some element are not found in either source or target"


class BackendConverter(Backend):
    """
    Convert two instance of backend to "align" them. This can be usefull if you have time series
    from a given "type of backend" (with certain names) but want to use another type 
    of backend to run the computation.
    
    Another usecase is if you have agents trained with certain backend type (so having loads, 
    generators etc. in certain order) but want to use a different backend to run the powerflow. You train
    your agent with a "simple and fast" powerflow but want to test them with a "more realistic" one.

    This means that grid2op will behave exactly as is the "source backend" class is used everywhere, but
    the powerflow computation will be carried out by the "target backend".
    
    # TODO: rename: "source backend" as "user backend" and "target backend" as "powerflow backend" or something
    like that.

    This means that from grid2op point of view, and from the agent point of view, line will be the order given
    by "source backend", load will be in the order of "source backend", topology will be given with the
    one from "source backend" etc. etc.

    Be careful, the BackendAction will also need to be transformed. Backend action is given with the order
    of the "source backend" and will have to be modified when transmitted to the "target backend".

    On the other end, no powerflow at all (except if some powerflows are performed at the initialization) will
    be computed using the source backend, only the target backend is relevant for the powerflow computations.

    Note that these backend need to access the grid description file from both "source backend" and "target backend"
    class. The underlying grid must be the same.

    # TODO: have a "mapping" from source name to target name in the constructor. If not provided it's the current
    behaviour (everything is automatic) but if present it maps the things according to this mapping. This mapping
    could be like `names_chronics_to_backend`
    
    Examples
    ---------
    Here is a (dummy and useless) example of how to use this class.

    .. code-block:: python

            import grid2op
            from grid2op.Converter import BackendConverter
            from grid2op.Backend import PandaPowerBackend
            from lightsim2grid import LightSimBackend
            backend = BackendConverter(source_backend_class=PandaPowerBackend,
                                       target_backend_class=LightSimBackend,
                                       target_backend_grid_path=None)

            # and now your environment behaves as if PandaPowerBackend did the computation (same load order, same
            generator order etc.) but real computation are made with LightSimBackend.
            # NB: for this specific example it is useless to do it because LightSimBackend and PandaPowerBackend have
            # by default the same order etc. This is just an illustration here

            # NB as of now you cannot use a runner with this method (yet)

            env = grid2op.make(..., backend=backend)

            # do regular computations here

    """
    IS_BK_CONVERTER = True
    
    def __init__(
        self,
        source_backend_class,
        target_backend_class,
        target_backend_grid_path=None,
        sub_source_target=None,
        detailed_infos_for_cascading_failures=False,
        use_target_backend_name=False,
        kwargs_target_backend=None,
        kwargs_source_backend=None,
    ):
        Backend.__init__(
            self,
            detailed_infos_for_cascading_failures=detailed_infos_for_cascading_failures,
            use_target_backend_name=use_target_backend_name,
            kwargs_target_backend=kwargs_target_backend,
            kwargs_source_backend=kwargs_source_backend,
        )
        difcf = detailed_infos_for_cascading_failures
        if kwargs_source_backend is None:
            kwargs_source_backend = {}
        self.source_backend = source_backend_class(
            detailed_infos_for_cascading_failures=difcf,
            **kwargs_source_backend
        )  # the one for the order of the elements
        if kwargs_target_backend is None:
            kwargs_target_backend = {}
        self.target_backend = target_backend_class(
            detailed_infos_for_cascading_failures=difcf,
            **kwargs_target_backend
        )  # the one to computes powerflow
        # if the target backend (the one performing the powerflows) needs a different file
        self.target_backend_grid_path = target_backend_grid_path

        # key: name in the source backend, value name in the target backend, for the substations
        self.sub_source_target = sub_source_target

        # if tmp is from the target backend, then tmp[self._line_tg2sr] is ordered according to the source backend
        self._sub_tg2sr = None
        self._sub_sr2tg = None
        self._line_tg2sr = None
        self._line_sr2tg = None
        self._gen_tg2sr = None
        self._gen_sr2tg = None
        self._load_tg2sr = None
        self._load_sr2tg = None
        self._shunt_tg2sr = None
        self._shunt_sr2tg = None
        self._topo_tg2sr = None
        self._topo_sr2tg = None
        self._storage_tg2sr = None
        self._storage_sr2tg = None

        # for redispatching data
        self.path_redisp = None
        self.name_redisp = None
        self.path_grid_layout = None
        self.name_grid_layout = None
        self.path_storage_data = None
        self.name_storage_data = None

        # for easier copy of np array
        self.cst1 = dt_float(1.0)

        # which name to use when loading the data (if names_chronics_to_backend has not been set)
        self.use_target_backend_name = use_target_backend_name 
        self.names_target_to_source = None
        
        # TODO storage check all this class ! + the doc of the backend

    def load_grid(self, path=None, filename=None):
        self.source_backend.load_grid(path, filename)
        # and now i load the target backend
        if self.target_backend_grid_path is not None:
            self.target_backend.load_grid(path=self.target_backend_grid_path)
        else:
            # both source and target backend understands the same format
            self.target_backend.load_grid(path, filename)

    def _assert_same_grid(self):
        """basic assertion that self and the target backend have the same grid
        but not necessarily the same object at the same place of course"""
        cls = type(self)
        tg_cls = type(self.target_backend)
        if cls.n_sub != tg_cls.n_sub:
            raise Grid2OpException(ERROR_NB_ELEMENTS.format("substations"))
        if cls.n_gen != tg_cls.n_gen:
            raise Grid2OpException(ERROR_NB_ELEMENTS.format("generators"))
        if cls.n_load != tg_cls.n_load:
            raise Grid2OpException(ERROR_NB_ELEMENTS.format("loads"))
        if cls.n_line != tg_cls.n_line:
            raise Grid2OpException(ERROR_NB_ELEMENTS.format("lines"))
        if cls.n_storage > 0:
            if cls.n_storage != tg_cls.n_storage:
                raise Grid2OpException(ERROR_NB_ELEMENTS.format("storages"))
        else:
            # a possible reason is that the "source backend" do not support storage
            # but the target one does. In this case I issue a warning,
            # and make sure I have not storage.
            if tg_cls.n_storage > 0:
                warnings.warn("BackendConverter: the source backend does not appear to support storage units, "
                              "but the target one does (and there are some storage units on the grid). "
                              "Be aware that the converted backend will NOT support storage units.")

    def _init_myself(self):
        # shortcut to set all information related to the class, except the name of the environment
        # this should been done when the source backend is fully initialized only
        self._assert_same_grid()

        # and now init all the converting vectors
        # a) for substation
        self._sub_tg2sr = np.full(self.n_sub, fill_value=-1, dtype=dt_int)
        self._sub_sr2tg = np.full(self.n_sub, fill_value=-1, dtype=dt_int)
        if self.sub_source_target is None:
            # automatic mode
            # I can only do it if the names matches
            if np.all(
                sorted(self.source_backend.name_sub)
                == sorted(self.target_backend.name_sub)
            ):
                for id_source, nm_source in enumerate(self.source_backend.name_sub):
                    id_target = np.where(self.target_backend.name_sub == nm_source)[0]
                    self._sub_tg2sr[id_source] = id_target
                    self._sub_sr2tg[id_target] = id_source
        else:
            for id_source, nm_source in enumerate(self.source_backend.name_sub):
                nm_target = self.sub_source_target[nm_source]
                id_target = np.where(self.target_backend.name_sub == nm_target)[0]
                self._sub_tg2sr[id_source] = id_target
                self._sub_sr2tg[id_target] = id_source

        # b) for load
        self._load_tg2sr = np.full(self.n_load, fill_value=-1, dtype=dt_int)
        self._load_sr2tg = np.full(self.n_load, fill_value=-1, dtype=dt_int)
        # automatic mode
        self._auto_fill_vect_load_gen_shunt(
            n_element=self.n_load,
            source_2_id_sub=self.source_backend.load_to_subid,
            target_2_id_sub=self.target_backend.load_to_subid,
            tg2sr=self._load_tg2sr,
            sr2tg=self._load_sr2tg,
            nm="load",
        )

        # c) for generator
        self._gen_tg2sr = np.full(self.n_gen, fill_value=-1, dtype=dt_int)
        self._gen_sr2tg = np.full(self.n_gen, fill_value=-1, dtype=dt_int)
        # automatic mode
        self._auto_fill_vect_load_gen_shunt(
            n_element=self.n_gen,
            source_2_id_sub=self.source_backend.gen_to_subid,
            target_2_id_sub=self.target_backend.gen_to_subid,
            tg2sr=self._gen_tg2sr,
            sr2tg=self._gen_sr2tg,
            nm="gen",
        )

        # d) for powerline
        self._line_tg2sr = np.full(self.n_line, fill_value=-1, dtype=dt_int)
        self._line_sr2tg = np.full(self.n_line, fill_value=-1, dtype=dt_int)
        # automatic
        self._auto_fill_vect_powerline()

        # e) and now the topology vectors.
        self._topo_tg2sr = np.full(self.source_backend.dim_topo, fill_value=-1, dtype=dt_int)
        self._topo_sr2tg = np.full(self.target_backend.dim_topo, fill_value=-1, dtype=dt_int)
        self._auto_fill_vect_topo()

        # f) for the storage units
        self._storage_tg2sr = np.full(self.n_storage, fill_value=-1, dtype=dt_int)
        self._storage_sr2tg = np.full(self.n_storage, fill_value=-1, dtype=dt_int)
        
        # automatic mode
        if self.n_storage:
            self._auto_fill_vect_load_gen_shunt(
                n_element=self.n_storage,
                source_2_id_sub=self.source_backend.storage_to_subid,
                target_2_id_sub=self.target_backend.storage_to_subid,
                tg2sr=self._storage_tg2sr,
                sr2tg=self._storage_sr2tg,
                nm="storage",
            )

        # shunt are available if both source and target provide it
        self.shunts_data_available = (
            self.source_backend.shunts_data_available
            and self.target_backend.shunts_data_available
        )
        if self.shunts_data_available:
            self._shunt_tg2sr = np.full(self.n_shunt, fill_value=-1, dtype=dt_int)
            self._shunt_sr2tg = np.full(self.n_shunt, fill_value=-1, dtype=dt_int)
            # automatic mode
            self._auto_fill_vect_load_gen_shunt(
                n_element=self.n_shunt,
                source_2_id_sub=self.source_backend.shunt_to_subid,
                target_2_id_sub=self.target_backend.shunt_to_subid,
                tg2sr=self._shunt_tg2sr,
                sr2tg=self._shunt_sr2tg,
                nm="shunt",
            )
        else:
            self.n_shunt = 0
            self.name_shunt = np.empty(0, dtype=str)
        self._thermal_limit_a = 1.0 * self.target_backend.thermal_limit_a
        self.set_thermal_limit(self.target_backend.thermal_limit_a[self._line_tg2sr])

    def _get_possible_target_ids(self, id_source, source_2_id_sub, target_2_id_sub, nm):
        id_sub_source = source_2_id_sub[id_source]
        id_sub_target = self._sub_tg2sr[id_sub_source]
        ids_target = np.where(target_2_id_sub == id_sub_target)[0]
        if ids_target.shape[0] == 0:
            raise RuntimeError(
                ERROR_ELEMENT_CONNECTED.format(nm, id_sub_target, id_sub_source)
            )
        return id_sub_target, ids_target

    def _auto_fill_vect_load_gen_shunt(
        self, n_element, source_2_id_sub, target_2_id_sub, tg2sr, sr2tg, nm
    ):
        nb_load_per_sub = np.zeros(self.n_sub, dtype=dt_int)
        if source_2_id_sub.shape[0] != n_element:
            raise RuntimeError(
                "Impossible to convert backend that do not have the same number of objects"
            )
        if target_2_id_sub.shape[0] != n_element:
            raise RuntimeError(
                "Impossible to convert backend that do not have the same number of objects"
            )
        for id_source in range(n_element):
            id_sub_target, id_target = self._get_possible_target_ids(
                id_source, source_2_id_sub, target_2_id_sub, nm
            )
            id_target = id_target[nb_load_per_sub[id_sub_target]]
            # TODO no no no use the "to_sub_pos" to compute that, and even better raise an error in this case
            # this means automatic is failing here !
            nb_load_per_sub[id_sub_target] += 1
            tg2sr[id_source] = id_target
            sr2tg[id_target] = id_source

    def _auto_fill_vect_powerline(self):
        # automatic matching
        nb_load_per_sub = np.zeros((self.n_sub, self.n_sub), dtype=dt_int)
        n_element = self.n_line
        source_or_2_id_sub = self.source_backend.line_or_to_subid
        target_or_2_id_sub = self.target_backend.line_or_to_subid
        source_ex_2_id_sub = self.source_backend.line_ex_to_subid
        target_ex_2_id_sub = self.target_backend.line_ex_to_subid
        nm = "powerline"
        tg2sr = self._line_tg2sr
        sr2tg = self._line_sr2tg
        for id_source in range(n_element):
            idor_sub_source = source_or_2_id_sub[id_source]
            idor_sub_target = self._sub_tg2sr[idor_sub_source]
            idex_sub_source = source_ex_2_id_sub[id_source]
            idex_sub_target = self._sub_tg2sr[idex_sub_source]
            ids_target = np.where(
                (target_or_2_id_sub == idor_sub_target)
                & (target_ex_2_id_sub == idex_sub_target)
            )[0]
            if ids_target.shape[0] == 0:
                raise RuntimeError(
                    ERROR_ELEMENT_CONNECTED.format(
                        nm,
                        "{}->{}".format(idor_sub_target, idex_sub_target),
                        "{}->{}".format(idor_sub_source, idex_sub_source),
                    )
                )
            id_target = ids_target[nb_load_per_sub[idor_sub_target, idex_sub_target]]
            # TODO no no no use the "to_sub_pos" to compute that, and even better raise an error in this case
            # this means automatic is failing here !
            nb_load_per_sub[idor_sub_target, idex_sub_target] += 1
            tg2sr[id_source] = id_target
            sr2tg[id_target] = id_source

    def _auto_fill_vect_topo(self):
        self._auto_fill_vect_topo_aux(
            self.n_load,
            self.source_backend.load_pos_topo_vect,
            self.target_backend.load_pos_topo_vect,
            self._load_sr2tg,
        )
        self._auto_fill_vect_topo_aux(
            self.n_gen,
            self.source_backend.gen_pos_topo_vect,
            self.target_backend.gen_pos_topo_vect,
            self._gen_sr2tg,
        )
        self._auto_fill_vect_topo_aux(
            self.n_line,
            self.source_backend.line_or_pos_topo_vect,
            self.target_backend.line_or_pos_topo_vect,
            self._line_sr2tg,
        )
        self._auto_fill_vect_topo_aux(
            self.n_line,
            self.source_backend.line_ex_pos_topo_vect,
            self.target_backend.line_ex_pos_topo_vect,
            self._line_sr2tg,
        )
        
        if self.n_storage:
            self._auto_fill_vect_topo_aux(
                self.n_storage,
                self.source_backend.storage_pos_topo_vect,
                self.target_backend.storage_pos_topo_vect,
                self._storage_sr2tg,
            )

    def _auto_fill_vect_topo_aux(self, n_elem, source_pos, target_pos, sr2tg):
        # TODO that might not be working as intented... it always says it's the identity...
        self._topo_tg2sr[source_pos[sr2tg]] = target_pos
        self._topo_sr2tg[target_pos] = source_pos[sr2tg]

    def assert_grid_correct(self):
        # this is done before a call to this function, by the environment
        tg_cls = type(self.target_backend)
        sr_cls = type(self.source_backend)
        
        env_name = type(self).env_name
        tg_cls.set_env_name(env_name)
        sr_cls.set_env_name(env_name)

        # handle specifc case of shunt data:
        if not self.target_backend.shunts_data_available:
            # disable the shunt data in grid2op.
            self.source_backend.shunts_data_available = False
            self.source_backend.n_shunt = None
            self.source_backend.name_shunt = np.empty(0, dtype=str)

        self._init_class_attr(obj=self.source_backend)
        if self.path_redisp is not None:
            # redispatching data were available
            try:
                super().load_redispacthing_data(self.path_redisp, name=self.name_redisp)
                self.source_backend.load_redispacthing_data(
                    self.path_redisp, name=self.name_redisp
                )
            except BackendError as exc_:
                self.redispatching_unit_commitment_availble = False
                warnings.warn(f"Impossible to load redispatching data. This is not an error but you will not be able "
                            f"to use all grid2op functionalities. "
                            f"The error was: \"{exc_}\"")
        if self.path_storage_data is not None:
            super().load_storage_data(self.path_storage_data, self.name_storage_data)
            self.source_backend.load_storage_data(
                self.path_storage_data, name=self.name_storage_data
            )
            self.target_backend.load_storage_data(
                self.path_storage_data, name=self.name_storage_data
            )
        if self.path_grid_layout is not None:
            # grid layout data were available
            super().load_grid_layout(self.path_grid_layout, self.name_grid_layout)
            self.source_backend.load_grid_layout(
                self.path_redisp, name=self.name_redisp
            )

        # init the target backend (the one that does the computation and that is initialized)
        self.target_backend.assert_grid_correct()

        # initialize the other one, because, well the grid should be seen from both backend
        self.source_backend._init_class_attr(obj=self)
        self.source_backend.assert_grid_correct()
        # and this should be called after all the rest
        super().assert_grid_correct()

        # everything went well, so i can properly terminate my initialization
        self._init_myself()
        
        # redefine this as the class changed after "assert grid correct"
        tg_cls = type(self.target_backend)
        sr_cls = type(self.source_backend)
        cls = type(self)

        if self.sub_source_target is None:
            # automatic mode for substations, names must match
            assert np.all(
                self.target_backend.name_sub[self._sub_tg2sr]
                == self.source_backend.name_sub
            )
            assert np.all(
                self.source_backend.name_sub[self._sub_sr2tg]
                == self.target_backend.name_sub
            )

        # check that all corresponding vectors are valid (and properly initialized, like every component above 0 etc.)
        self._check_both_consistent(self._line_tg2sr, self._line_sr2tg)
        self._check_both_consistent(self._load_tg2sr, self._load_sr2tg)
        self._check_both_consistent(self._gen_tg2sr, self._gen_sr2tg)
        self._check_both_consistent(self._sub_tg2sr, self._sub_sr2tg)
        
        if cls.n_storage == tg_cls.n_storage:
            # both source and target supports storage units
            self._check_both_consistent(self._topo_tg2sr, self._topo_sr2tg)
        elif self.n_storage == 0:
            # n_storage == 0 and there are storage units on the source backend
            # this means that the target_backend supports storage but not
            # the source one
            assert np.all(self._topo_sr2tg[self._topo_tg2sr] >= 0)
            assert np.all(sorted(self._topo_sr2tg[self._topo_tg2sr]) == np.arange(self.dim_topo))
            
            topo_sr2tg_without_storage = self._topo_sr2tg[self._topo_sr2tg >= 0]
            assert (self._topo_sr2tg == -1).sum() == tg_cls.n_storage
            assert np.all(self._topo_tg2sr[topo_sr2tg_without_storage] >= 0)
            target_without_storage = np.array([i for i in range(tg_cls.dim_topo) 
                                               if not i in tg_cls.storage_pos_topo_vect])
            assert np.all(sorted(self._topo_tg2sr[topo_sr2tg_without_storage]) == target_without_storage)
            self._topo_sr2tg = topo_sr2tg_without_storage

        if self.shunts_data_available:
            self._check_both_consistent(self._shunt_tg2sr, self._shunt_sr2tg)

        # finally check that powergrids are identical (up to the env name)
        tg_cls.same_grid_class(sr_cls)
        
        # once everything is done, make the converter for the names
        d_loads = {tg_cls.name_load[i]: sr_cls.name_load[self._load_sr2tg[i]]
                    for i in range(cls.n_load)}
        d_gens = {tg_cls.name_gen[i]: sr_cls.name_gen[self._gen_sr2tg[i]]
                    for i in range(cls.n_gen)}
        d_lines = {tg_cls.name_line[i]: sr_cls.name_line[self._line_sr2tg[i]]
                    for i in range(cls.n_line)}
        d_subs = {tg_cls.name_sub[i]: sr_cls.name_sub[self._sub_sr2tg[i]]
                    for i in range(cls.n_sub)}
        dict_ = {"loads": d_loads, "lines": d_lines, "prods": d_gens, "subs": d_subs}
        self.names_target_to_source = dict_

    def _check_vect_valid(self, vect):
        assert np.all(
            vect >= 0
        ), ERROR_INVALID_VECTOR
        assert sorted(np.unique(vect)) == sorted(
            vect
        ), ERROR_INVALID_VECTOR
        if vect.shape[0] > 0:
            assert (
                np.max(vect) == vect.shape[0] - 1
            ), ERROR_INVALID_VECTOR

    def _check_both_consistent(self, tg2sr, sr2tg):
        self._check_vect_valid(tg2sr)
        self._check_vect_valid(sr2tg)
        res = np.arange(tg2sr.shape[0])
        assert np.all(tg2sr[sr2tg] == res)
        assert np.all(sr2tg[tg2sr] == res)

    def assert_grid_correct_after_powerflow(self):
        # we don't assert that `self.source_backend.assert_grid_correct_after_powerflow()`
        # because obviously no powerflow are run using the source backend.
        self.target_backend.assert_grid_correct_after_powerflow()
        super().assert_grid_correct_after_powerflow()
        self._sh_vnkv = self.target_backend._sh_vnkv

    def reset(self, grid_path, grid_filename=None):
        """
        Reload the power grid.
        For backwards compatibility this method calls `Backend.load_grid`.
        But it is encouraged to overload it in the subclasses.
        """
        self.target_backend.reset(grid_path, grid_filename=grid_filename)

    def close(self):
        self.source_backend.close()
        self.target_backend.close()

    def apply_action(self, action):
        # action is from the source backend
        action_target = self._transform_action(action)
        self.target_backend.apply_action(action_target)

    def runpf(self, is_dc=False):
        return self.target_backend.runpf(is_dc=is_dc)

    def copy(self):
        source_backend_sv = self.source_backend
        target_backend_sv = self.target_backend
        self.source_backend = None
        self.target_backend = None
        res = copy.deepcopy(self)
        res.source_backend = source_backend_sv.copy()
        res.target_backend = target_backend_sv.copy()
        self.source_backend = source_backend_sv
        self.target_backend = target_backend_sv
        return res

    def save_file(self, full_path):
        self.target_backend.save_file(full_path)
        self.source_backend.save_file(full_path)

    def get_line_status(self):
        tmp = self.target_backend.get_line_status()
        return tmp[self._line_tg2sr]

    def get_line_flow(self):
        tmp = self.target_backend.get_line_flow()
        return self.cst1 * tmp[self._line_tg2sr]

    def set_thermal_limit(self, limits):
        super().set_thermal_limit(limits=limits)
        self.source_backend.set_thermal_limit(limits=limits)
        if limits is not None:
            self.target_backend.set_thermal_limit(limits=limits[self._line_sr2tg])

    def get_thermal_limit(self):
        tmp = self.target_backend.get_thermal_limit()
        return self.cst1 * tmp[self._line_tg2sr]

    def get_topo_vect(self):
        tmp = self.target_backend.get_topo_vect()
        return tmp[self._topo_tg2sr]

    def generators_info(self):
        prod_p, prod_q, prod_v = self.target_backend.generators_info()
        return (
            self.cst1 * prod_p[self._gen_tg2sr],
            self.cst1 * prod_q[self._gen_tg2sr],
            self.cst1 * prod_v[self._gen_tg2sr],
        )

    def loads_info(self):
        load_p, load_q, load_v = self.target_backend.loads_info()
        return (
            self.cst1 * load_p[self._load_tg2sr],
            self.cst1 * load_q[self._load_tg2sr],
            self.cst1 * load_v[self._load_tg2sr],
        )

    def lines_or_info(self):
        p_, q_, v_, a_ = self.target_backend.lines_or_info()
        return (
            self.cst1 * p_[self._line_tg2sr],
            self.cst1 * q_[self._line_tg2sr],
            self.cst1 * v_[self._line_tg2sr],
            self.cst1 * a_[self._line_tg2sr],
        )

    def lines_ex_info(self):
        p_, q_, v_, a_ = self.target_backend.lines_ex_info()
        return (
            self.cst1 * p_[self._line_tg2sr],
            self.cst1 * q_[self._line_tg2sr],
            self.cst1 * v_[self._line_tg2sr],
            self.cst1 * a_[self._line_tg2sr],
        )

    def storages_info(self):
        p_, q_, v_ = self.target_backend.storages_info()
        return (
            self.cst1 * p_[self._storage_tg2sr],
            self.cst1 * q_[self._storage_tg2sr],
            self.cst1 * v_[self._storage_tg2sr],
        )

    def shunt_info(self):
        if self._shunt_tg2sr is not None:
            # shunts are supported by both source and target backend
            sh_p, sh_q, sh_v, sh_bus = self.target_backend.shunt_info()
            return (
                sh_p[self._shunt_tg2sr],
                sh_q[self._shunt_tg2sr],
                sh_v[self._shunt_tg2sr],
                sh_bus[self._shunt_tg2sr],
            )
        # shunt are not supported by either source or target backend
        return [], [], [], []

    def sub_from_bus_id(self, bus_id):
        # not supported because the bus_id is given into the source backend,
        # and i need to convert to to the target backend, not sure how to do that atm
        raise Grid2OpException(
            "This backend doesn't allow to get the substation from the bus id."
        )

    def _disconnect_line(self, id_):
        id_target = int(
            self._line_tg2sr[id_]
        )  # not sure why, but it looks to work this way
        self.target_backend._disconnect_line(id_target)

    def _transform_action(self, source_action):
        # transform the source action into the target backend action
        # source_action: a backend action!
        target_action = copy.deepcopy(source_action)
        # consistent with TestLoadingBackendFunc, otherwise it's not correct
        target_action.reorder(
            no_load=self._load_sr2tg,
            no_gen=self._gen_sr2tg,
            no_topo=self._topo_sr2tg,
            no_shunt=self._shunt_sr2tg,
            no_storage=self._storage_sr2tg,
        )
        return target_action

    def load_redispacthing_data(self, path, name="prods_charac.csv"):
        # data are loaded with the name of the source backend, i need to map it to the target backend too
        self.path_redisp = path
        self.name_redisp = name

    def load_storage_data(self, path, name="storage_units_charac.csv"):
        # data are loaded with the name of the source backend, i need to map it to the target backend too
        self.path_storage_data = path
        self.name_storage_data = name

    def load_grid_layout(self, path, name="grid_layout.json"):
        self.path_grid_layout = path
        self.name_grid_layout = name

    def get_action_to_set(self):
        act = self.target_backend.get_action_to_set()
        line_vect = self._line_sr2tg
        gen_vect = self._gen_sr2tg
        load_vect = self._load_sr2tg
        topo_vect = self._topo_sr2tg
        dict_ = act._dict_inj
        if "prod_p" in dict_:
            dict_["dict_"] = dict_["prod_p"][gen_vect]
        if "prod_v" in dict_:
            dict_["dict_"] = dict_["prod_v"][gen_vect]
        if "load_p" in dict_:
            dict_["dict_"] = dict_["load_p"][load_vect]
        if "load_q" in dict_:
            dict_["dict_"] = dict_["load_q"][load_vect]

        act._set_topo_vect[:] = act._set_topo_vect[topo_vect]
        act._change_bus_vect[:] = act._change_bus_vect[topo_vect]
        act._hazards[:] = act._hazards[line_vect]
        act._maintenance[:] = act._hazards[line_vect]
        act._redispatch[:] = act._redispatch[gen_vect]
        act._set_line_status[:] = act._set_line_status[line_vect]
        act._switch_line_status[:] = act._switch_line_status[line_vect]

        if act.shunt_added and act.shunts_data_available:
            shunt_vect = self._shunt_sr2tg
            act.shunt_p[:] = act.shunt_p[shunt_vect]
            act.shunt_q[:] = act.shunt_q[shunt_vect]
            act.shunt_bus[:] = act.shunt_bus[shunt_vect]

        return act

    def update_thermal_limit(self, env):
        # TODO
        # env has the powerline stored in the order of the source backend, but i need
        # to have them stored in the order of the target backend for such function
        pass

    # TODO update_from_obs too, maybe ?



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from typing import Tuple
from grid2op.Action import BaseAction
from grid2op.Observation import BaseObservation
from grid2op.Environment.environment import Environment


class _ForecastEnv(Environment):
    """Type of environment that increments the `highres_simulator` when it calls the env.step method.
    
    It is used by obs.get_forecast_env.
    """
    def __init__(self, *args, **kwargs):
        if "_update_obs_after_reward" not in kwargs:
            kwargs["_update_obs_after_reward"] = False
        super().__init__(*args, **kwargs)
        
    def step(self, action: BaseAction) -> Tuple[BaseObservation, float, bool, dict]:
        self._highres_sim_counter += 1
        return super().step(action)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from typing import Tuple
import copy
import warnings
import numpy as np

from grid2op.dtypes import dt_int, dt_bool, dt_float
from grid2op.Observation import ObservationSpace
from grid2op.Exceptions import Grid2OpException

from grid2op.gym_compat.utils import (_compute_extra_power_for_losses,
                                      GYM_AVAILABLE,
                                      GYMNASIUM_AVAILABLE,
                                      check_gym_version)

ALL_ATTR_OBS = (
    "year",
    "month",
    "day",
    "hour_of_day",
    "minute_of_hour",
    "day_of_week",
    "gen_p",
    "gen_p_before_curtail",
    "gen_q",
    "gen_v",
    "gen_margin_up",
    "gen_margin_down",
    "load_p",
    "load_q",
    "load_v",
    "p_or",
    "q_or",
    "v_or",
    "a_or",
    "p_ex",
    "q_ex",
    "v_ex",
    "a_ex",
    "rho",
    "line_status",
    "timestep_overflow",
    "topo_vect",
    "time_before_cooldown_line",
    "time_before_cooldown_sub",
    "time_next_maintenance",
    "duration_next_maintenance",
    "target_dispatch",
    "actual_dispatch",
    "storage_charge",
    "storage_power_target",
    "storage_power",
    "curtailment",
    "curtailment_limit",
    "curtailment_limit_effective",
    "thermal_limit",
    "is_alarm_illegal",
    "time_since_last_alarm",
    "last_alarm",
    "attention_budget",
    "was_alarm_used_after_game_over",
    "max_step",
    "active_alert",
    "attack_under_alert",
    "time_since_last_alert",
    "alert_duration",
    "total_number_of_alert",
    "time_since_last_attack",
    "was_alert_used_after_attack",
    "theta_or",
    "theta_ex",
    "load_theta",
    "gen_theta",
)

# TODO add the alarm stuff
# TODO add the time step
# TODO add the is_illegal and co there


class __AuxBoxGymObsSpace:
    """
    This class allows to convert a grid2op observation space into a gym "Box" which is
    a regular Box in R^d.

    It also allows to customize which part of the observation you want to use and offer capacity to
    center / reduce the data or to use more complex function from the observation.
    
    .. warning::
        Depending on the presence absence of gymnasium and gym packages this class might behave differently.
        
        In grid2op we tried to maintain compatibility both with gymnasium (newest) and gym (legacy, 
        no more maintained) RL packages. The behaviour is the following:
        
        - :class:`BoxGymObsSpace` will inherit from gymnasium if it's installed 
          (in this case it will be :class:`BoxGymnasiumObsSpace`), otherwise it will
          inherit from gym (and will be exactly :class:`BoxLegacyGymObsSpace`)
        - :class:`BoxGymnasiumObsSpace` will inherit from gymnasium if it's available and never from
          from gym
        - :class:`BoxLegacyGymObsSpace` will inherit from gym if it's available and never from
          from gymnasium
        
        See :ref:`gymnasium_gym` for more information
    
    .. note::
        A gymnasium Box is encoded as a numpy array.
        
    Examples
    --------
    If you simply want to use it you can do:

    .. code-block:: python

        import grid2op
        env_name = "l2rpn_case14_sandbox"  # or any other name
        env = grid2op.make(env_name)

        from grid2op.gym_compat import GymEnv, BoxGymObsSpace
        gym_env = GymEnv(env)

        gym_env.observation_space = BoxGymObsSpace(env.observation_space)

    In this case it will extract all the features in all the observation (a detailed list is given
    in the documentation at :ref:`observation_module`.

    You can select the attribute you want to keep, for example:

    .. code-block:: python

        gym_env.observation_space = BoxGymObsSpace(env.observation_space,
                                                   attr_to_keep=['load_p', "gen_p", "rho])

    You can also apply some basic transformation to the attribute of the observation before building
    the resulting gym observation (which in this case is a vector). This can be done with:

    .. code-block:: python

        gym_env.observation_space = BoxGymObsSpace(env.observation_space,
                                                   attr_to_keep=['load_p', "gen_p", "rho"],
                                                   divide={"gen_p": env.gen_pmax},
                                                   substract={"gen_p": 0.5 * env.gen_pmax})

    In the above example, the resulting "gen_p" part of the vector will be given by the following
    formula: `gym_obs = (grid2op_obs - substract) / divide`.

    Hint: you can use: divide being the standard deviation and subtract being the average of the attribute
    on a few episodes for example. This can be done with :class:`grid2op.utils.EpisodeStatistics` for example.

    Finally, you can also modify more the attribute of the observation and add it to your box. This
    can be done rather easily with the "functs" argument like:

    .. code-block:: python

        gym_env.observation_space = BoxGymObsSpace(env.observation_space,
                                                   attr_to_keep=["connectivity_matrix", "log_load"],
                                                   functs={"connectivity_matrix":
                                                              (lambda grid2opobs: grid2opobs.connectivity_matrix().flatten(),
                                                               0., 1.0, None, None),
                                                           "log_load":
                                                            (lambda grid2opobs: np.log(grid2opobs.load_p),
                                                            None, 10., None, None)
                                                        }
                                                   )

    In this case, "functs" should be a dictionary, the "keys" should be string (keys should also be
    present in the `attr_to_keep` list) and the values should count 5 elements
    (callable, low, high, shape, dtype) with:

    - `callable` a function taking as input a grid2op observation and returning a numpy array
    - `low` (optional) (put None if you don't want to specify it, defaults to `-np.inf`) the lowest value
      your numpy array can take. It can be a single number or an array with the same shape
      as the return value of your function.
    - `high` (optional) (put None if you don't want to specify it, defaults to `np.inf`) the highest value
      your numpy array can take. It can be a single number or an array with the same shape
      as the return value of your function.
    - `shape` (optional) (put None if you don't want to specify it) the shape of the return value
      of your function. It should be a tuple (and not a single number). By default it is computed
      with by applying your function to an observation.
    - `dtype` (optional, put None if you don't want to change it, defaults to np.float32) the type of
      the numpy array as output of your function.

    Notes
    -----
    The range of the values for "gen_p" / "prod_p" are not strictly `env.gen_pmin` and `env.gen_pmax`.
    This is due to the "approximation" when some redispatching is performed (the precision of the
    algorithm that computes the actual dispatch from the information it receives) and also because
    sometimes the losses of the grid are really different that the one anticipated in the "chronics" (yes
    env.gen_pmin and env.gen_pmax are not always ensured in grid2op)

    """

    def __init__(
        self,
        grid2op_observation_space,
        attr_to_keep=ALL_ATTR_OBS,
        subtract=None,
        divide=None,
        functs=None,
    ):
        check_gym_version(type(self)._gymnasium)
        if not isinstance(grid2op_observation_space, ObservationSpace):
            raise RuntimeError(
                f"Impossible to create a BoxGymObsSpace without providing a "
                f"grid2op observation. You provided {type(grid2op_observation_space)}"
                f'as the "grid2op_observation_space" attribute.'
            )
        self._attr_to_keep = sorted(attr_to_keep)

        ob_sp = grid2op_observation_space
        tol_redisp = (
            ob_sp.obs_env._tol_poly
        )  # add to gen_p otherwise ... well it can crash
        extra_for_losses = _compute_extra_power_for_losses(ob_sp)
        
        self._dict_properties = {
            "year": (
                np.zeros(1, dtype=dt_int),
                np.zeros(1, dtype=dt_int) + 2200,
                (1,),
                dt_int,
            ),
            "month": (
                np.zeros(1, dtype=dt_int),
                np.zeros(1, dtype=dt_int) + 12,
                (1,),
                dt_int,
            ),
            "day": (
                np.zeros(1, dtype=dt_int),
                np.zeros(1, dtype=dt_int) + 31,
                (1,),
                dt_int,
            ),
            "hour_of_day": (
                np.zeros(1, dtype=dt_int),
                np.zeros(1, dtype=dt_int) + 24,
                (1,),
                dt_int,
            ),
            "minute_of_hour": (
                np.zeros(1, dtype=dt_int),
                np.zeros(1, dtype=dt_int) + 60,
                (1,),
                dt_int,
            ),
            "day_of_week": (
                np.zeros(1, dtype=dt_int),
                np.zeros(1, dtype=dt_int) + 7,
                (1,),
                dt_int,
            ),
            "current_step": (
                np.zeros(1, dtype=dt_int),
                np.zeros(1, dtype=dt_int) + np.iinfo(dt_int).max,
                (1,),
                dt_int,
            ),
            "gen_p": (
                np.full(shape=(ob_sp.n_gen,), fill_value=0.0, dtype=dt_float)
                - tol_redisp
                - extra_for_losses,
                ob_sp.gen_pmax + tol_redisp + extra_for_losses,
                (ob_sp.n_gen,),
                dt_float,
            ),
            "gen_q": (
                np.full(shape=(ob_sp.n_gen,), fill_value=-np.inf, dtype=dt_float),
                np.full(shape=(ob_sp.n_gen,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_gen,),
                dt_float,
            ),
            "gen_v": (
                np.full(shape=(ob_sp.n_gen,), fill_value=0.0, dtype=dt_float),
                np.full(shape=(ob_sp.n_gen,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_gen,),
                dt_float,
            ),
            "gen_margin_up": (
                np.full(shape=(ob_sp.n_gen,), fill_value=0.0, dtype=dt_float),
                1.0 * ob_sp.gen_max_ramp_up,
                (ob_sp.n_gen,),
                dt_float,
            ),
            "gen_margin_down": (
                np.full(shape=(ob_sp.n_gen,), fill_value=0.0, dtype=dt_float),
                1.0 * ob_sp.gen_max_ramp_down,
                (ob_sp.n_gen,),
                dt_float,
            ),
            "gen_theta": (
                np.full(shape=(ob_sp.n_gen,), fill_value=-180., dtype=dt_float),
                np.full(shape=(ob_sp.n_gen,), fill_value=180., dtype=dt_float),
                (ob_sp.n_gen,),
                dt_float,
            ),
            "load_p": (
                np.full(shape=(ob_sp.n_load,), fill_value=-np.inf, dtype=dt_float),
                np.full(shape=(ob_sp.n_load,), fill_value=+np.inf, dtype=dt_float),
                (ob_sp.n_load,),
                dt_float,
            ),
            "load_q": (
                np.full(shape=(ob_sp.n_load,), fill_value=-np.inf, dtype=dt_float),
                np.full(shape=(ob_sp.n_load,), fill_value=+np.inf, dtype=dt_float),
                (ob_sp.n_load,),
                dt_float,
            ),
            "load_v": (
                np.full(shape=(ob_sp.n_load,), fill_value=0.0, dtype=dt_float),
                np.full(shape=(ob_sp.n_load,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_load,),
                dt_float,
            ),
            "load_theta": (
                np.full(shape=(ob_sp.n_load,), fill_value=-180., dtype=dt_float),
                np.full(shape=(ob_sp.n_load,), fill_value=180., dtype=dt_float),
                (ob_sp.n_load,),
                dt_float,
            ),
            "p_or": (
                np.full(shape=(ob_sp.n_line,), fill_value=-np.inf, dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "q_or": (
                np.full(shape=(ob_sp.n_line,), fill_value=-np.inf, dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "a_or": (
                np.full(shape=(ob_sp.n_line,), fill_value=0.0, dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "v_or": (
                np.full(shape=(ob_sp.n_line,), fill_value=0.0, dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "theta_or": (
                np.full(shape=(ob_sp.n_line,), fill_value=-180., dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=180., dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "p_ex": (
                np.full(shape=(ob_sp.n_line,), fill_value=-np.inf, dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "q_ex": (
                np.full(shape=(ob_sp.n_line,), fill_value=-np.inf, dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "a_ex": (
                np.full(shape=(ob_sp.n_line,), fill_value=0.0, dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "v_ex": (
                np.full(shape=(ob_sp.n_line,), fill_value=0.0, dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "theta_ex": (
                np.full(shape=(ob_sp.n_line,), fill_value=-180., dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=180., dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "rho": (
                np.full(shape=(ob_sp.n_line,), fill_value=0.0, dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "line_status": (
                np.full(shape=(ob_sp.n_line,), fill_value=0, dtype=dt_int),
                np.full(shape=(ob_sp.n_line,), fill_value=1, dtype=dt_int),
                (ob_sp.n_line,),
                dt_int,
            ),
            "timestep_overflow": (
                np.full(
                    shape=(ob_sp.n_line,), fill_value=np.iinfo(dt_int).min, dtype=dt_int
                ),
                np.full(
                    shape=(ob_sp.n_line,), fill_value=np.iinfo(dt_int).max, dtype=dt_int
                ),
                (ob_sp.n_line,),
                dt_int,
            ),
            "topo_vect": (
                np.full(shape=(ob_sp.dim_topo,), fill_value=-1, dtype=dt_int),
                np.full(shape=(ob_sp.dim_topo,), fill_value=2, dtype=dt_int),
                (ob_sp.dim_topo,),
                dt_int,
            ),
            "time_before_cooldown_line": (
                np.full(shape=(ob_sp.n_line,), fill_value=0, dtype=dt_int),
                np.full(
                    shape=(ob_sp.n_line,), fill_value=np.iinfo(dt_int).max, dtype=dt_int
                ),
                (ob_sp.n_line,),
                dt_int,
            ),
            "time_before_cooldown_sub": (
                np.full(shape=(ob_sp.n_sub,), fill_value=0, dtype=dt_int),
                np.full(
                    shape=(ob_sp.n_sub,), fill_value=np.iinfo(dt_int).max, dtype=dt_int
                ),
                (ob_sp.n_sub,),
                dt_int,
            ),
            "time_next_maintenance": (
                np.full(shape=(ob_sp.n_line,), fill_value=-1, dtype=dt_int),
                np.full(
                    shape=(ob_sp.n_line,), fill_value=np.iinfo(dt_int).max, dtype=dt_int
                ),
                (ob_sp.n_line,),
                dt_int,
            ),
            "duration_next_maintenance": (
                np.full(shape=(ob_sp.n_line,), fill_value=0, dtype=dt_int),
                np.full(
                    shape=(ob_sp.n_line,), fill_value=np.iinfo(dt_int).max, dtype=dt_int
                ),
                (ob_sp.n_line,),
                dt_int,
            ),
            "target_dispatch": (
                np.minimum(ob_sp.gen_pmin, -ob_sp.gen_pmax),
                np.maximum(-ob_sp.gen_pmin, +ob_sp.gen_pmax),
                (ob_sp.n_gen,),
                dt_float,
            ),
            "actual_dispatch": (
                np.minimum(ob_sp.gen_pmin, -ob_sp.gen_pmax),
                np.maximum(-ob_sp.gen_pmin, +ob_sp.gen_pmax),
                (ob_sp.n_gen,),
                dt_float,
            ),
            "storage_charge": (
                np.full(shape=(ob_sp.n_storage,), fill_value=0, dtype=dt_float),
                1.0 * ob_sp.storage_Emax,
                (ob_sp.n_storage,),
                dt_float,
            ),
            "storage_power_target": (
                -1.0 * ob_sp.storage_max_p_prod,
                1.0 * ob_sp.storage_max_p_absorb,
                (ob_sp.n_storage,),
                dt_float,
            ),
            "storage_power": (
                -1.0 * ob_sp.storage_max_p_prod,
                1.0 * ob_sp.storage_max_p_absorb,
                (ob_sp.n_storage,),
                dt_float,
            ),
            "storage_theta": (
                np.full(shape=(ob_sp.n_storage,), fill_value=-180., dtype=dt_float),
                np.full(shape=(ob_sp.n_storage,), fill_value=180., dtype=dt_float),
                (ob_sp.n_storage,),
                dt_float,
            ),
            "curtailment": (
                np.full(shape=(ob_sp.n_gen,), fill_value=0.0, dtype=dt_float),
                np.full(shape=(ob_sp.n_gen,), fill_value=1.0, dtype=dt_float),
                (ob_sp.n_gen,),
                dt_float,
            ),
            "curtailment_limit": (
                np.full(shape=(ob_sp.n_gen,), fill_value=0.0, dtype=dt_float),
                np.full(shape=(ob_sp.n_gen,), fill_value=1.0, dtype=dt_float),
                (ob_sp.n_gen,),
                dt_float,
            ),
            "curtailment_mw": (
                np.full(shape=(ob_sp.n_gen,), fill_value=0.0, dtype=dt_float),
                1.0 * ob_sp.gen_pmax,
                (ob_sp.n_gen,),
                dt_float,
            ),
            "curtailment_limit_mw": (
                np.full(shape=(ob_sp.n_gen,), fill_value=0.0, dtype=dt_float),
                1.0 * ob_sp.gen_pmax,
                (ob_sp.n_gen,),
                dt_float,
            ),
            "thermal_limit": (
                np.full(shape=(ob_sp.n_line,), fill_value=0.0, dtype=dt_float),
                np.full(shape=(ob_sp.n_line,), fill_value=np.inf, dtype=dt_float),
                (ob_sp.n_line,),
                dt_float,
            ),
            "is_alarm_illegal": (
                np.full(shape=(1,), fill_value=False, dtype=dt_bool),
                np.full(shape=(1,), fill_value=True, dtype=dt_bool),
                (1,),
                dt_bool,
            ),
            "time_since_last_alarm": (
                np.full(shape=(1,), fill_value=-1, dtype=dt_int),
                np.full(shape=(1,), fill_value=np.iinfo(dt_int).max, dtype=dt_int),
                (1,),
                dt_int,
            ),
            "last_alarm": (
                np.full(shape=(ob_sp.dim_alarms,), fill_value=-1, dtype=dt_int),
                np.full(
                    shape=(ob_sp.dim_alarms,),
                    fill_value=np.iinfo(dt_int).max,
                    dtype=dt_int,
                ),
                (ob_sp.dim_alarms,),
                dt_int,
            ),
            "attention_budget": (
                np.full(shape=(1,), fill_value=-1, dtype=dt_float),
                np.full(shape=(1,), fill_value=np.inf, dtype=dt_float),
                (1,),
                dt_float,
            ),
            "was_alarm_used_after_game_over": (
                np.full(shape=(1,), fill_value=False, dtype=dt_bool),
                np.full(shape=(1,), fill_value=True, dtype=dt_bool),
                (1,),
                dt_bool,
            ),
            "delta_time": (
                np.full(shape=(1,), fill_value=0, dtype=dt_float),
                np.full(shape=(1,), fill_value=np.inf, dtype=dt_float),
                (1,),
                dt_float,
            ),
            # alert stuff
            "active_alert": (
                np.full(shape=(ob_sp.dim_alerts,), fill_value=False, dtype=dt_bool),
                np.full(shape=(ob_sp.dim_alerts,), fill_value=True, dtype=dt_bool),
                (ob_sp.dim_alerts,),
                dt_bool,
            ),
            "time_since_last_alert": (
                np.full(shape=(ob_sp.dim_alerts,), fill_value=-1, dtype=dt_int),
                np.full(shape=(ob_sp.dim_alerts,), fill_value=np.iinfo(dt_int).max, dtype=dt_int),
                (ob_sp.dim_alerts,),
                dt_int,
            ),
            "alert_duration": (
                np.full(shape=(ob_sp.dim_alerts,), fill_value=-1, dtype=dt_int),
                np.full(shape=(ob_sp.dim_alerts,), fill_value=np.iinfo(dt_int).max, dtype=dt_int),
                (ob_sp.dim_alerts,),
                dt_int,
            ),
            "total_number_of_alert": (
                np.full(shape=(1 if ob_sp.dim_alerts else 0,), fill_value=-1, dtype=dt_int),
                np.full(shape=(1 if ob_sp.dim_alerts else 0,), fill_value=np.iinfo(dt_int).max, dtype=dt_int),
                (1 if ob_sp.dim_alerts else 0,),
                dt_int,
            ),
            "time_since_last_attack": (
                np.full(shape=(ob_sp.dim_alerts,), fill_value=-1, dtype=dt_int),
                np.full(shape=(ob_sp.dim_alerts,), fill_value=np.iinfo(dt_int).max, dtype=dt_int),
                (ob_sp.dim_alerts,),
                dt_int,
            ),
            "was_alert_used_after_attack": (
                np.full(shape=(ob_sp.dim_alerts,), fill_value=-1, dtype=dt_int),
                np.full(shape=(ob_sp.dim_alerts,), fill_value=1, dtype=dt_int),
                (ob_sp.dim_alerts,),
                dt_int,
            ),
            "attack_under_alert": (
                np.full(shape=(ob_sp.dim_alerts,), fill_value=-1, dtype=dt_int),
                np.full(shape=(ob_sp.dim_alerts,), fill_value=1, dtype=dt_int),
                (ob_sp.dim_alerts,),
                dt_int,
            ),
        }
        self._dict_properties["max_step"] = copy.deepcopy(self._dict_properties["current_step"])
        self._dict_properties["delta_time"] = copy.deepcopy(self._dict_properties["current_step"])
        self._dict_properties["prod_p"] = copy.deepcopy(self._dict_properties["gen_p"])
        self._dict_properties["prod_q"] = copy.deepcopy(self._dict_properties["gen_q"])
        self._dict_properties["prod_v"] = copy.deepcopy(self._dict_properties["gen_v"])
        self._dict_properties["gen_p_before_curtail"] = copy.deepcopy(self._dict_properties["gen_p"])
        self._dict_properties["curtailment_limit_effective"] = copy.deepcopy(self._dict_properties[
            "curtailment_limit"
        ])
        
        if functs is None:
            functs = {}
            
        for key in functs.keys():
            if key not in self._attr_to_keep:
                raise RuntimeError(
                    f'The key {key} is present in the "functs" dictionary but not in the '
                    f'"attr_to_keep". This is not consistent: either ignore this function, '
                    f'in that case remove "{key}" from "functs" or you want to add '
                    f'something to your observation, in that case add it to "attr_to_keep"'
                )

        if subtract is None:
            subtract = {}
        self._subtract = subtract
        if divide is None:
            divide = {}
        self._divide = divide

        # handle the "functional" part
        self._template_obs = ob_sp._template_obj.copy()
        self.__func = {}

        self._dims = None
        low, high, shape, dtype = self._get_info(functs)

        # initialize the base container
        type(self)._BoxType.__init__(self, low=low, high=high, shape=shape, dtype=dtype)
        
        # convert data in `_subtract` and `_divide` to the right type
        self._fix_value_sub_div(self._subtract, functs)
        self._fix_value_sub_div(self._divide, functs)

    def _get_shape(self, el, functs):
        if el in functs:
            callable_, low_, high_, shape_, dtype_ = functs[el]
        elif el in self._dict_properties:
            # el is an attribute of an observation, for example "load_q" or "topo_vect"
            low_, high_, shape_, dtype_ = self._dict_properties[el]
        return shape_
    
    def _fix_value_sub_div(self, dict_, functs):
        """dict_ is either self._subtract or self._divide"""
        keys = list(dict_.keys())
        for k in keys:
            v = dict_[k]
            if isinstance(v, (list, tuple)):
                v = np.array(v).astype(self.dtype)
            else:
                shape = self._get_shape(k, functs)
                v = np.full(shape, fill_value=v, dtype=self.dtype)
            dict_[k] = v
        
    def _get_info(self, functs):
        low = None
        high = None
        shape = None
        dtype = None
        self._dims = []
        for el in self._attr_to_keep:
            if el in functs:
                # the attribute name "el" has been put in the functs
                try:
                    callable_, low_, high_, shape_, dtype_ = functs[el]
                except Exception as exc_:
                    raise RuntimeError(
                        f'When using keyword argument "functs" you need to provide something '
                        f"like: (callable_, low_, high_, shape_, dtype_) for each key. "
                        f'There was an error with "{el}".'
                        f"The error was:\n {exc_}"
                    )

                try:
                    tmp = callable_(self._template_obs.copy())
                except Exception as exc_:
                    raise RuntimeError(
                        f'Error for the function your provided with key "{el}" (using the'
                        f'"functs" dictionary) '
                        f"The error was :\n {exc_}"
                    )
                if not isinstance(tmp, np.ndarray):
                    raise RuntimeError(
                        f'The result of the function you provided as part of the "functs"'
                        f"dictionary for key {el}"
                        f"do not return a numpy array. This is not supported."
                    )
                self.__func[el] = callable_
                if dtype_ is None:
                    dtype_ = dt_float
                if shape_ is None:
                    shape_ = tmp.shape

                if not isinstance(shape_, tuple):
                    raise RuntimeError(
                        "You need to provide a tuple as a shape of the output of your data"
                    )

                if low_ is None:
                    low_ = np.full(shape_, fill_value=-np.inf, dtype=dtype_)
                elif isinstance(low_, float):
                    low_ = np.full(shape_, fill_value=low_, dtype=dtype_)

                if high_ is None:
                    high_ = np.full(shape_, fill_value=np.inf, dtype=dtype_)
                elif isinstance(high_, float):
                    high_ = np.full(shape_, fill_value=high_, dtype=dtype_)

                if ((tmp < low_) | (tmp > high_)).any():
                    raise RuntimeError(
                        f"Wrong value for low / high in the functs argument for key {el}. Please"
                        f"fix the low_ / high_ in the tuple ( callable_, low_, high_, shape_, dtype_)."
                    )

            elif el in self._dict_properties:
                # el is an attribute of an observation, for example "load_q" or "topo_vect"
                low_, high_, shape_, dtype_ = self._dict_properties[el]
            else:
                li_keys = "\n\t-".join(
                    sorted(list(self._dict_properties.keys()) + list(self.__func.keys()))
                )
                raise RuntimeError(
                    f'Unknown observation attributes "{el}". Supported attributes are: '
                    f"\n{li_keys}"
                )

            # handle the data type
            if dtype is None:
                dtype = dtype_
            else:
                if dtype_ == dt_float:
                    # promote whatever to float anyway
                    dtype = dt_float
                elif dtype_ == dt_int and dtype == dt_bool:
                    # promote bool to int
                    dtype = dt_int

            # handle the shape
            if shape is None:
                shape = shape_
            else:
                shape = (shape[0] + shape_[0],)

            # handle low / high
            if el in self._subtract:
                low_ =  1.0 * low_.astype(dtype)
                high_ =  1.0 * high_.astype(dtype)
                low_ -= self._subtract[el]
                high_ -= self._subtract[el]
                
            if el in self._divide:
                low_ =  1.0 * low_.astype(dtype)
                high_ =  1.0 * high_.astype(dtype)
                low_ /= self._divide[el]
                high_ /= self._divide[el]
            if low is None:
                low = low_
                high = high_
            else:
                low = np.concatenate((low.astype(dtype), low_.astype(dtype))).astype(
                    dtype
                )
                high = np.concatenate((high.astype(dtype), high_.astype(dtype))).astype(
                    dtype
                )

            # remember where this need to be stored
            self._dims.append(shape[0])

        return low, high, shape, dtype

    def _handle_attribute(self, grid2op_observation, attr_nm):
        res = getattr(grid2op_observation, attr_nm).astype(self.dtype)
        if attr_nm in self._subtract:
            res -= self._subtract[attr_nm]
        if attr_nm in self._divide:
            res /= self._divide[attr_nm]
        return res

    def to_gym(self, grid2op_observation):
        """
        This is the function that is called to transform a grid2Op observation, sent by the grid2op environment
        and convert it to a numpy array (an element of a gym Box)

        Parameters
        ----------
        grid2op_observation:
            The grid2op observation (as a grid2op object)

        Returns
        -------
        res: :class:`numpy.ndarray`
            A numpy array compatible with the openAI gym Box that represents the action space.

        """
        res = np.empty(shape=self.shape, dtype=self.dtype)
        prev = 0
        for attr_nm, where_to_put in zip(self._attr_to_keep, self._dims):
            if attr_nm in self.__func:
                tmp = self.__func[attr_nm](grid2op_observation)
            elif hasattr(grid2op_observation, attr_nm):
                tmp = self._handle_attribute(grid2op_observation, attr_nm)
            else:
                raise RuntimeError(f'Unknown attribute "{attr_nm}".')
            res[prev:where_to_put] = tmp
            prev = where_to_put
        return res

    def close(self):
        pass

    def get_indexes(self, key: str) -> Tuple[int, int]:
        """Allows to retrieve the indexes of the gym action that
        are concerned by the attribute name `key` given in input.

        .. versionadded:: 1.9.3
        
        .. warning::
            Copy paste from box_gym_act_space, need refacto !
            
        Parameters
        ----------
        key : str
            the attribute name (*eg* "set_storage" or "redispatch")

        Returns
        -------
        Tuple[int, int]
            _description_

        Examples
        --------
        
        You can use it like:
        
        .. code-block:: python
        
            gym_env = ... # an environment with a BoxActSpace
            
            act = np.zeros(gym_env.action_space.shape)
            key = "redispatch"  # "redispatch", "curtail", "set_storage"
            start_, end_ = gym_env.action_space.get_indexes(key)
            act[start_:end_] = np.random.uniform(high=1, low=-1, size=env.gen_redispatchable.sum())
            # act only modifies the redispatch with the input given (here a uniform redispatching between -1 and 1)
            
        """
        error_msg =(f"Impossible to use the grid2op action property \"{key}\""
                    f"with this action space.")
        if key not in self._attr_to_keep:
            raise Grid2OpException(error_msg)
        prev = 0
        for attr_nm, where_to_put in zip(
            self._attr_to_keep, self._dims
        ):
            if attr_nm == key:
                return prev, where_to_put
            prev = where_to_put
        raise Grid2OpException(error_msg)
    
    def normalize_attr(self, attr_nm: str):
        """
        This function normalizes the part of the space
        that corresponds to the attribute `attr_nm`.

        The normalization consists in having a vector between 0. and 1.
        It is achieved by:
        
        - dividing by the range (high - low)
        - adding the minimum value (low).

        .. note::
            It only affects continuous attribute. No error / warnings are
            raised if you attempt to use it on a discrete attribute.

        .. warning::
            This normalization relies on the `high` and `low` attribute. It cannot be done if
            the attribute is not bounded (for example when its maximum limit is `np.inf`). A warning
            is raised in this case.
            
        Parameters
        ----------
        attr_nm : `str`
            The name of the attribute to normalize
        """
        if attr_nm in self._divide or attr_nm in self._subtract:
            raise Grid2OpException(
                f"Cannot normalize attribute \"{attr_nm}\" that you already "
                f"modified with either `divide` or `subtract` (observation space)."
            )
        prev = 0
        if self.dtype != dt_float:
            raise Grid2OpException(
                "Cannot normalize attribute with a observation "
                "space that is not float !"
            )
        for attr_tmp, where_to_put in zip(self._attr_to_keep, self._dims):
            if attr_tmp == attr_nm:
                curr_high = 1.0 * self.high[prev:where_to_put]
                curr_low = 1.0 * self.low[prev:where_to_put]
                finite_high = np.isfinite(curr_high)
                finite_low = np.isfinite(curr_high)
                both_finite = finite_high & finite_low
                both_finite &= curr_high > curr_low

                if (~both_finite).any():
                    warnings.warn(f"The normalization of attribute \"{both_finite}\" cannot be performed entirely as "
                                  f"there are some non finite value, or `high == `low` "
                                  f"for some components.")
                    
                self._divide[attr_nm] = np.ones(curr_high.shape, dtype=self.dtype)
                self._subtract[attr_nm] = np.zeros(curr_high.shape, dtype=self.dtype)

                self._divide[attr_nm][both_finite] = (
                    curr_high[both_finite] - curr_low[both_finite]
                )
                self._subtract[attr_nm][both_finite] += curr_low[both_finite]

                self.high[prev:where_to_put][both_finite] = 1.0
                self.low[prev:where_to_put][both_finite] = 0.0
                break
            prev = where_to_put


if GYM_AVAILABLE:
    from gym.spaces import Box as LegGymBox
    from grid2op.gym_compat.base_gym_attr_converter import BaseLegacyGymAttrConverter
    BoxLegacyGymObsSpace = type("BoxLegacyGymObsSpace",
                                (__AuxBoxGymObsSpace, LegGymBox, ),
                                {"_gymnasium": False,
                                 "_BaseGymAttrConverterType": BaseLegacyGymAttrConverter,
                                 "_BoxType": LegGymBox,
                                 "__module__": __name__})
    BoxLegacyGymObsSpace.__doc__ = __AuxBoxGymObsSpace.__doc__
    BoxGymObsSpace = BoxLegacyGymObsSpace
    BoxGymObsSpace.__doc__ = __AuxBoxGymObsSpace.__doc__
        

if GYMNASIUM_AVAILABLE:
    from gymnasium.spaces import Box
    from grid2op.gym_compat.base_gym_attr_converter import BaseGymnasiumAttrConverter
    BoxGymnasiumObsSpace = type("BoxGymnasiumObsSpace",
                                (__AuxBoxGymObsSpace, Box, ),
                                {"_gymnasium": True,
                                 "_BaseGymAttrConverterType": BaseGymnasiumAttrConverter,
                                 "_BoxType": Box,
                                 "__module__": __name__})
    BoxGymnasiumObsSpace.__doc__ = __AuxBoxGymObsSpace.__doc__
    BoxGymObsSpace = BoxGymnasiumObsSpace
    BoxGymObsSpace.__doc__ = __AuxBoxGymObsSpace.__doc__



__all__ = [
    "handlers",
    "ChronicsHandler",
    "GridValue",
    "ChangeNothing",
    "Multifolder",
    "MultifolderWithCache",
    "GridStateFromFile",
    "GridStateFromFileWithForecasts",
    "GridStateFromFileWithForecastsWithMaintenance",
    "GridStateFromFileWithForecastsWithoutMaintenance",
    "FromNPY",
    "FromChronix2grid",
    "FromHandlers"
]

from grid2op.Chronics.chronicsHandler import ChronicsHandler
from grid2op.Chronics.changeNothing import ChangeNothing
from grid2op.Chronics.gridValue import GridValue
from grid2op.Chronics.gridStateFromFile import GridStateFromFile
from grid2op.Chronics.gridStateFromFileWithForecasts import (
    GridStateFromFileWithForecasts,
)
from grid2op.Chronics.multiFolder import Multifolder
from grid2op.Chronics.GSFFWFWM import GridStateFromFileWithForecastsWithMaintenance
from grid2op.Chronics.fromFileWithoutMaintenance import (
    GridStateFromFileWithForecastsWithoutMaintenance,
)
from grid2op.Chronics.multifolderWithCache import MultifolderWithCache
from grid2op.Chronics.fromNPY import FromNPY
from grid2op.Chronics.fromChronix2grid import FromChronix2grid
from grid2op.Chronics.time_series_from_handlers import FromHandlers



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Converter import IdToAct
from grid2op.Agent.agentWithConverter import AgentWithConverter


class RandomAgent(AgentWithConverter):
    """
    This agent acts randomly on the powergrid. It uses the :class:`grid2op.Converters.IdToAct` to compute all the
    possible actions available for the environment. And then chooses a random one among all these.

    Notes
    ------
    Actions are taken uniformly at random among unary actions. For example, if a game rules allows to take actions that
    can disconnect a powerline AND modify the topology of a substation an action that do both will not be sampled
    by this class.

    This agent is not equivalent to calling `env.action_space.sample()` because the sampling is not
    done the same manner. This agent sample uniformly among all unary actions whereas
    `env.action_space.sample()` (see :func:`grid2op.Action.SerializableActionSpace.sample` for more
    information about the later).

    """

    def __init__(
        self, action_space, action_space_converter=IdToAct, **kwargs_converter
    ):
        AgentWithConverter.__init__(
            self, action_space, action_space_converter, **kwargs_converter
        )

    def my_act(self, transformed_observation, reward, done=False):
        """
        A random agent will "simply" draw a random number between 0 and the number of action, and return this action.

        This is equivalent to draw uniformly at random a feasible action.

        Notes
        -----
        In order to be working as intended, it is crucial that this method does not rely on any other source
        of "pseudo randomness" than :attr:`grid2op.Space.RandomObject.space_prng`.

        In particular, you must avoid
        to use `np.random.XXXX` or the `random` python module. You can replace any call to `np.random.XXX` by
        `self.space_prng.XXX` (**eg**  `np.random.randint(1,5)` can be replaced by `self.space_prng.randint(1,5)`).

        If you really need other sources of randomness (for example if you use tensorflow or torch) we strongly
        recommend you to overload the :func:`BaseAgent.seed` accordingly.

        """
        my_int = self.space_prng.randint(self.action_space.n)
        return my_int



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Reward.baseReward import BaseReward
from grid2op.dtypes import dt_float


class ConstantReward(BaseReward):
    """
    Most basic implementation of reward: everything has the same values: 0.0

    Note that this :class:`BaseReward` subtype is not useful at all, whether to train an :attr:`BaseAgent`
    nor to assess its performance of course.


    Examples
    ---------
    You can use this reward in any environment with:

    .. code-block:

        import grid2op
        from grid2op.Reward import ConstantReward

        # then you create your environment with it:
        NAME_OF_THE_ENVIRONMENT = "rte_case14_realistic"
        env = grid2op.make(NAME_OF_THE_ENVIRONMENT,reward_class=ConstantReward)
        # and do a step with a "do nothing" action
        obs = env.reset()
        obs, reward, done, info = env.step(env.action_space())
        # the reward is 0., always... Not really useful

    """

    def __init__(self, logger=None):
        BaseReward.__init__(self, logger=logger)

    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        return dt_float(0.0)



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import copy

from grid2op.Reward.baseReward import BaseReward
from grid2op.Reward.constantReward import ConstantReward
from grid2op.Exceptions import Grid2OpException


class RewardHelper:
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\
        It is a class internal to the :class:`grid2op.Environment.Environment` do not use outside
        of its purpose and do not attempt to modify it.

    This class aims at making the creation of rewards class more automatic by the :class:`grid2op.Environment`.

    It is not recommended to derived or modified this class. If a different reward need to be used, it is recommended
    to build another object of this class, and change the :attr:`RewardHelper.rewardClass` attribute.

    Attributes
    ----------
    rewardClass: ``type``
        Type of reward that will be use by this helper. Note that the type (and not an instance / object of that type)
        must be given here. It defaults to :class:`ConstantReward`

    template_reward: :class:`BaseReward`
        An object of class :attr:`RewardHelper.rewardClass` used to compute the rewards.

    """

    def __init__(self, reward_func=ConstantReward, logger=None):
        self.rewardClass = None
        self.template_reward = None
        self.logger = logger
        self.change_reward(reward_func)

    def initialize(self, env):
        """
        This function initializes the template_reward with the environment. It is used especially for using
        :func:`RewardHelper.range`.

        Parameters
        ----------
        env: :class:`grid2op.Environment.Environment`
            The current used environment.

        """
        self.template_reward.initialize(env)

    def range(self):
        """
        Provides the range of the rewards.

        Returns
        -------
        res: ``(float, float)``
            The minimum reward per time step (possibly infinity) and the maximum reward per timestep (possibly infinity)
        """
        return self.template_reward.get_range()

    def reset(self, env):
        """called each time `env.reset` is invoked"""
        self.template_reward.reset(env)

    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        """
        Gives the reward that follows the execution of the :class:`grid2op.BaseAction.BaseAction` action in the
        :class:`grid2op.Environment.Environment` env;

        Parameters
        ----------
        action: :class:`grid2op.Action.Action`
            The action performed by the BaseAgent.

        env: :class:`grid2op.Environment.Environment`
            The current environment.

        has_error: ``bool``
            Does the action caused an error, such a diverging powerflow for example= (``True``: the action caused
            an error)

        is_done: ``bool``
            Is the game over (``True`` = the game is over)

        is_illegal: ``bool``
            Is the action legal or not (``True`` = the action was illegal). See
            :class:`grid2op.Exceptions.IllegalAction` for more information.

        is_ambiguous: ``bool``
            Is the action ambiguous or not (``True`` = the action was ambiguous). See
            :class:`grid2op.Exceptions.AmbiguousAction` for more information.

        Returns
        -------
        res: ``float``
            The computed reward

        """
        res = self.template_reward(
            action, env, has_error, is_done, is_illegal, is_ambiguous
        )
        return res

    def change_reward(self, reward_func):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

            Use `env.change_reward` instead (:func:`grid2op.Environment.BaseEnv.change_reward`)

        """
        if isinstance(reward_func, BaseReward):
            # reward object given directly
            self.rewardClass = type(reward_func)
            self.template_reward = copy.deepcopy(reward_func)
        elif issubclass(reward_func, BaseReward):
            # reward is provided as a class
            self.rewardClass = reward_func
            try:
                self.template_reward = reward_func(logger=self.logger)
            except TypeError as exc_:
                self.logger.warn(f"Reward \"{reward_func.__name__}\" does not support the logger feature. Error was : {exc_}")
                # old (<= 1.7.0) behaviour
                self.template_reward = reward_func()
        else:
            raise Grid2OpException(
                f"Impossible to build a reward with input reward_func={reward_func}. "
                f"NB `reward_func` should be either an object of type `BaseReward` (or "
                f"one of its derivative) "
                f"or a class that inherit from `BaseReward`"
            )

    def close(self):
        """clsoe the reward helper (in case there are specific behaviour for certain rewards"""
        self.template_reward.close()



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.
import numpy as np
import warnings
from datetime import datetime, timedelta
from abc import ABC, abstractmethod

from grid2op.dtypes import dt_int
from grid2op.Space import RandomObject
from grid2op.Exceptions import EnvError

# TODO sous echantillonner ou sur echantilloner les scenario: need to modify everything that affect the number
# TODO of time steps there, for example "Space.gen_min_time_on" or "params.NB_TIMESTEP_POWERFLOW_ALLOWED" for
# TODO example. And more generally, it would be better to have all of this attributes exported / imported in
# TODO time interval, instead of time steps.

# TODO add a class to sample "online" the data.


class GridValue(RandomObject, ABC):
    """
    This is the base class for every kind of data for the _grid.

    It allows the :class:`grid2op.Environment` to perform powergrid modification that make the "game" time dependant.

    It is not recommended to directly create :class:`GridValue` object, but to use the
    :attr:`grid2op.Environment.chronics_handler" for such a purpose. This is made in an attempt to make sure the
    :func:`GridValue.initialize` is called. Before this initialization, it is not recommended to use any
    :class:`GridValue` object.

    The method :func:`GridValue.next_chronics` should be used between two epoch of the game. If there are no more
    data to be generated from this object, then :func:`GridValue.load_next` should raise a :class:`StopIteration`
    exception and a call to :func:`GridValue.done` should return True.

    In grid2op, the production and loads (and hazards or maintenance) can be stored in this type of
    of "GridValue". This class will map things generated (or read from a file) and assign the given element
    of the powergrid with its proper value at each time steps.

    Attributes
    ----------
    time_interval: :class:`.datetime.timedelta`
        Time interval between 2 consecutive timestamps. Default 5 minutes.

    start_datetime:  :class:`datetime.datetime`
        The datetime of the first timestamp of the scenario.

    current_datetime: :class:`datetime.datetime`
        The timestamp of the current scenario.

    max_iter: ``int``
        Number maximum of data to generate for one episode.

    curr_iter: ``int``
        Duration of the current episode.

    maintenance_time: ``numpy.ndarray``, dtype:``int``
        Number of time steps the next maintenance will take place with the following convention:

            - -1 no maintenance are planned for the forseeable future
            - 0 a maintenance is taking place
            - 1, 2, 3 ... a maintenance will take place in 1, 2, 3, ... time step

        Some examples are given in :func:`GridValue.maintenance_time_1d`.

    maintenance_duration: ``numpy.ndarray``, dtype:``int``
        Duration of the next maintenance. 0 means no maintenance is happening. If a maintenance is planned for a
        given powerline, this number decreases each time step, up until arriving at 0 when the maintenance is over. Note
        that if a maintenance is planned (see :attr:`GridValue.maintenance_time`) this number indicates how long
        the maintenance will last, and does not suppose anything on the maintenance taking place or not (= there can be
        positive number here without a powerline being removed from the grid for maintenance reason). Some examples are
        given in :func:`GridValue.maintenance_duration_1d`.

    hazard_duration: ``numpy.ndarray``, dtype:``int``
        Duration of the next hzard. 0 means no maintenance is happening. If a hazard is taking place for a
        given powerline, this number decreases each time step, up until arriving at 0 when the maintenance is over. On
        the contrary to :attr:`GridValue.maintenance_duration`, if a component of this vector is higher than 1, it
        means that the powerline is out of service. Some examples are
        given in :func:`GridValue.get_hazard_duration_1d`.


    
    """

    NAN_BUT_IN_INT = -9999999
    
    def __init__(
        self,
        time_interval=timedelta(minutes=5),
        max_iter=-1,
        start_datetime=datetime(year=2019, month=1, day=1),
        chunk_size=None,
    ):
        RandomObject.__init__(self)

        self.time_interval = time_interval
        self.current_datetime = start_datetime
        self.start_datetime = start_datetime
        self._max_iter = max_iter
        self.curr_iter = 0

        self.maintenance_time = None
        self.maintenance_duration = None
        self.hazard_duration = None

    def get_kwargs(self, dict_):
        """
        Overload this function if you want to pass some data when building
        a new instance of this class.
        """
        pass

    @property
    def max_iter(self):
        return self._max_iter
    
    @max_iter.setter
    def max_iter(self, value : int):
        self._max_iter = int(value)
    
    @abstractmethod
    def initialize(
        self,
        order_backend_loads,
        order_backend_prods,
        order_backend_lines,
        order_backend_subs,
        names_chronics_to_backend,
    ):
        """
        This function is used to initialize the data generator.
        It can be use to load scenarios, or to initialize noise if scenarios are generated on the fly. It must also
        initialize :attr:`GridValue.maintenance_time`, :attr:`GridValue.maintenance_duration` and
        :attr:`GridValue.hazard_duration`.

        This function should also increment :attr:`GridValue.curr_iter` of 1 each time it is called.

        The :class:`GridValue` is what makes the connection between the data (generally in a shape of files on the
        hard drive) and the power grid. One of the main advantage of the Grid2Op package is its ability to change
        the tool that computes the load flows. Generally, such :class:`grid2op.Backend` expects data in a specific
        format that is given by the way their internal powergrid is represented, and in particular, the "same"
        objects can have different name and different position. To ensure that the same chronics would
        produce the same results on every backend (**ie** regardless of the order of which the Backend is expecting
        the data, the outcome of the powerflow is the same) we encourage the user to provide a file that maps the name
        of the object in the chronics to the name of the same object in the backend.

        This is done with the "names_chronics_to_backend" dictionnary that has the following keys:

          - "loads"
          - "prods"
          - "lines"

        The value associated to each of these keys is in turn a mapping dictionnary from the chronics to the backend.
        This means that each *keys* of these subdictionnary is a name of one column in the files, and each values
        is the corresponding name of this same object in the dictionnary. An example is provided bellow.

        Parameters
        ----------
        order_backend_loads: ``numpy.ndarray``, dtype:str
            Ordered name, in the Backend, of the loads. It is required that a :class:`grid2op.Backend` object always
            output the informations in the same order. This array gives the name of the loads following this order.
            See the documentation of :mod:`grid2op.Backend` for more information about this.

        order_backend_prods: ``numpy.ndarray``, dtype:str
            Same as order_backend_loads, but for generators.

        order_backend_lines: ``numpy.ndarray``, dtype:str
            Same as order_backend_loads, but for powerline.

        order_backend_subs: ``numpy.ndarray``, dtype:str
            Same as order_backend_loads, but for powerline.

        names_chronics_to_backend: ``dict``
            See in the description of the method for more information about its format.

        Examples
        --------
        For example, suppose we have a :class:`grid2op.Backend` with:

          - substations ids strart from 0 to N-1 (N being the number of substations in the powergrid)
          - loads named "load_i" with "i" the subtations to which it is connected
          - generators units named "gen_i" (i still the substation id to which it is connected)
          - powerlnes are named "i_j" if it connected substations i to substation j

        And on the other side, we have some files with the following conventions:

          - substations are numbered from 1 to N
          - loads are named "i_C" with i being the substation to which it is connected
          - generators are named "i_G" with is being the id of the substations to which it is connected
          - powerlines are namesd "i_j_k" where i is the origin substation, j the extremity substations and "k"
            is a unique identifier of this powerline in the powergrid.

        In this case, instead of renaming the powergrid (in the backend) of the data files, it is advised to build the
        following elements and initialize the object gridval of type :class:`GridValue` with:

        .. code-block:: python

            gridval = GridValue()  # Note: this code won't execute because "GridValue" is an abstract class
            order_backend_loads = ['load_1', 'load_2', 'load_13', 'load_3', 'load_4', 'load_5', 'load_8', 'load_9',
                                     'load_10', 'load_11', 'load_12']
            order_backend_prods = ['gen_1', 'gen_2', 'gen_5', 'gen_7', 'gen_0']
            order_backend_lines = ['0_1', '0_4', '8_9', '8_13', '9_10', '11_12', '12_13', '1_2', '1_3', '1_4', '2_3',
                                       '3_4', '5_10', '5_11', '5_12', '3_6', '3_8', '4_5', '6_7', '6_8']
            order_backend_subs = ['sub_0', 'sub_1', 'sub_10', 'sub_11', 'sub_12', 'sub_13', 'sub_2', 'sub_3', 'sub_4',
                                      'sub_5', 'sub_6', 'sub_7', 'sub_8', 'sub_9']
            names_chronics_to_backend = {"loads": {"2_C": 'load_1', "3_C": 'load_2',
                                                       "14": 'load_13', "4_C": 'load_3', "5_C": 'load_4',
                                                       "6_C": 'load_5', "9_C": 'load_8', "10_C": 'load_9',
                                                       "11_C": 'load_10', "12_C": 'load_11',
                                                       "13_C": 'load_12'},
                                             "lines": {'1_2_1': '0_1', '1_5_2': '0_4', '9_10_16': '8_9', '9_14_17': '8_13',
                                                      '10_11_18': '9_10', '12_13_19': '11_12', '13_14_20': '12_13',
                                                       '2_3_3': '1_2', '2_4_4': '1_3', '2_5_5': '1_4', '3_4_6': '2_3',
                                                       '4_5_7': '3_4', '6_11_11': '5_10', '6_12_12': '5_11',
                                                       '6_13_13': '5_12', '4_7_8': '3_6', '4_9_9': '3_8', '5_6_10': '4_5',
                                                      '7_8_14': '6_7', '7_9_15': '6_8'},
                                             "prods": {"1_G": 'gen_0', "3_G": "gen_2", "6_G": "gen_5",
                                                       "2_G": "gen_1", "8_G": "gen_7"},
                                            }
            gridval.initialize(order_backend_loads, order_backend_prods, order_backend_lines, names_chronics_to_backend)

        """
        self.curr_iter += 1
        self.current_datetime += self.time_interval

    @staticmethod
    def get_maintenance_time_1d(maintenance):
        """
        This function allows to transform a 1d numpy aarray maintenance, where is specify:

            - 0 there is no maintenance at this time step
            - 1 there is a maintenance at this time step

        Into the representation in terms of "next maintenance time" as specified in
        :attr:`GridValue.maintenance_time` which is:

            - `-1` no foreseeable maintenance operation will be performed
            - `0` a maintenance operation is being performed
            - `1`, `2` etc. is the number of time step the next maintenance will be performed.

        Parameters
        ----------
        maintenance: ``numpy.ndarray``
            1 dimensional array representing the time series of the maintenance (0 there is no maintenance, 1 there
            is a maintenance at this time step)

        Returns
        -------
        maintenance_duration: ``numpy.ndarray``
            Array representing the time series of the duration of the next maintenance forseeable.

        Examples
        --------

        If no maintenance are planned:

        .. code-block:: python

            maintenance_time = GridValue.get_maintenance_time_1d(np.array([0 for _ in range(10)]))
            assert np.all(maintenance_time == np.array([-1  for _ in range(10)]))


        If a maintenance planned of 3 time steps starting at timestep 6 (index 5 - index starts at 0)

        .. code-block:: python

            maintenance = np.array([0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0])
            maintenance_time = GridValue.get_maintenance_time_1d(maintenance)
            assert np.all(maintenance_time == np.array([5,4,3,2,1,0,0,0,-1,-1,-1,-1,-1,-1,-1,-1]))

        If a maintenance planned of 3 time steps starting at timestep 6
        (index 5 - index starts at 0), and a second one for 2 time steps at time step 13

        .. code-block:: python

            maintenance = np.array([0,0,0,0,0,1,1,1,0,0,0,0,1,1,0,0,0])
            maintenance_time = GridValue.get_maintenance_time_1d(maintenance)
            assert np.all(maintenance_time == np.array([5,4,3,2,1,0,0,0,4,3,2,1,0,0,-1,-1,-1]))

        """

        res = np.full(maintenance.shape, fill_value=GridValue.NAN_BUT_IN_INT, dtype=dt_int)
        maintenance = np.concatenate((maintenance, (0, 0)))
        a = np.diff(maintenance)
        # +1 is because numpy does the diff `t+1` - `t` so to get index of the initial array
        # I need to "+1"
        start = np.where(a == 1)[0] + 1  # start of maintenance
        end = np.where(a == -1)[0] + 1  # end of maintenance
        prev_ = 0
        # it's efficient here as i do a loop only on the number of time there is a maintenance
        # and maintenance are quite rare
        for beg_, end_ in zip(start, end):
            res[prev_:beg_] = list(range(beg_ - prev_, 0, -1))
            res[beg_:end_] = 0
            prev_ = end_

        # no maintenance are planned in the forseeable future
        res[prev_:] = -1
        return res

    @staticmethod
    def get_maintenance_duration_1d(maintenance):
        """
        This function allows to transform a 1d numpy aarray maintenance (or hazards), where is specify:

            - 0 there is no maintenance at this time step
            - 1 there is a maintenance at this time step

        Into the representation in terms of "next maintenance duration" as specified in
        :attr:`GridValue.maintenance_duration` which is:

            - `0` no forseeable maintenance operation will be performed
            - `1`, `2` etc. is the number of time step the next maintenance will last (it can be positive even in the
                case that no maintenance is currently being performed.

        Parameters
        ----------
        maintenance: ``numpy.ndarray``
            1 dimensional array representing the time series of the maintenance (0 there is no maintenance, 1 there
            is a maintenance at this time step)

        Returns
        -------
        maintenance_duration: ``numpy.ndarray``
            Array representing the time series of the duration of the next maintenance forseeable.

        Examples
        --------

        If no maintenance are planned:

        .. code-block:: python

            maintenance = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
            maintenance_duration = GridValue.get_maintenance_duration_1d(maintenance)
            assert np.all(maintenance_duration == np.array([0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]))

        If a maintenance planned of 3 time steps starting at timestep 6 (index 5 - index starts at 0)

        .. code-block:: python

            maintenance = np.array([0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0])
            maintenance_duration = GridValue.get_maintenance_duration_1d(maintenance)
            assert np.all(maintenance_duration == np.array([3,3,3,3,3,3,2,1,0,0,0,0,0,0,0,0]))

        If a maintenance planned of 3 time steps starting at timestep 6
        (index 5 - index starts at 0), and a second one for 2 time steps at time step 13

        .. code-block:: python

            maintenance = np.array([0,0,0,0,0,1,1,1,0,0,0,0,1,1,0,0,0])
            maintenance_duration = GridValue.get_maintenance_duration_1d(maintenance)
            assert np.all(maintenance_duration == np.array([3,3,3,3,3,3,2,1,2,2,2,2,2,1,0,0,0]))

        """
        res = np.full(maintenance.shape, fill_value=GridValue.NAN_BUT_IN_INT, dtype=dt_int)
        maintenance = np.concatenate((maintenance, (0, 0)))
        a = np.diff(maintenance)
        # +1 is because numpy does the diff `t+1` - `t` so to get index of the initial array
        # I need to "+1"
        start = np.where(a == 1)[0] + 1  # start of maintenance
        end = np.where(a == -1)[0] + 1  # end of maintenance
        prev_ = 0
        # it's efficient here as i do a loop only on the number of time there is a maintenance
        # and maintenance are quite rare
        for beg_, end_ in zip(start, end):
            res[prev_:beg_] = end_ - beg_
            res[beg_:end_] = list(range(end_ - beg_, 0, -1))
            prev_ = end_

        # no maintenance are planned in the foreseeable future
        res[prev_:] = 0
        return res

    @staticmethod
    def get_hazard_duration_1d(hazard):
        """
        This function allows to transform a 1d numpy aarray maintenance (or hazards), where is specify:

            - 0 there is no maintenance at this time step
            - 1 there is a maintenance at this time step

        Into the representation in terms of "hzard duration" as specified in
        :attr:`GridValue.maintenance_duration` which is:

            - `0` no forseeable hazard operation will be performed
            - `1`, `2` etc. is the number of time step the next hzard will last (it is positive only when a hazard
                affect a given powerline)

        Compared to :func:`GridValue.get_maintenance_duration_1d` we only know when the hazard occurs how long it
        will last.

        Parameters
        ----------
        hazard: ``numpy.ndarray``
            1 dimensional array representing the time series of the hazards (0 there is no hazard, 1 there
            is a hazard at this time step)

        Returns
        -------
        hazard_duration: ``numpy.ndarray``
            Array representing the time series of the duration of the next hazard forseeable.

        Examples
        --------

        If no maintenance are planned:

        .. code-block:: python

            hazard = np.array([0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0])
            hazard_duration = GridValue.get_hazard_duration_1d(hazard)
            assert np.all(hazard_duration == np.array([0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]))

        If a maintenance planned of 3 time steps starting at timestep 6 (index 5 - index starts at 0)

        .. code-block:: python

            hazard = np.array([0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,0])
            hazard_duration = GridValue.get_hazard_duration_1d(hazard)
            assert np.all(hazard_duration == np.array([0,0,0,0,0,3,2,1,0,0,0,0,0,0,0,0]))

        If a maintenance planned of 3 time steps starting at timestep 6
        (index 5 - index starts at 0), and a second one for 2 time steps at time step 13

        .. code-block:: python

            hazard = np.array([0,0,0,0,0,1,1,1,0,0,0,0,1,1,0,0,0])
            hazard_duration = GridValue.get_hazard_duration_1d(hazard)
            assert np.all(hazard_duration == np.array([0,0,0,0,0,3,2,1,0,0,0,0,2,1,0,0,0]))

        """

        res = np.full(hazard.shape, fill_value=GridValue.NAN_BUT_IN_INT, dtype=dt_int)
        hazard = np.concatenate((hazard, (0, 0)))
        a = np.diff(hazard)
        # +1 is because numpy does the diff `t+1` - `t` so to get index of the initial array
        # I need to "+1"
        start = np.where(a == 1)[0] + 1  # start of maintenance
        end = np.where(a == -1)[0] + 1  # end of maintenance
        prev_ = 0
        # it's efficient here as i do a loop only on the number of time there is a maintenance
        # and maintenance are quite rare
        for beg_, end_ in zip(start, end):
            res[prev_:beg_] = 0
            res[(beg_):(end_)] = list(range(end_ - beg_, 0, -1))
            prev_ = end_

        # no maintenance are planned in the forseeable future
        res[prev_:] = 0
        return res

    @abstractmethod
    def load_next(self):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

            This is automatically called by the "env.step" function. It loads the next information
            about the grid state (load p and load q, prod p and prod v as well as some maintenance
            or hazards information)

        Generate the next values, either by reading from a file, or by generating on the fly and return a dictionary
        compatible with the :class:`grid2op.BaseAction` class allowed for the :class:`Environment`.

        More information about this dictionary can be found at :func:`grid2op.BaseAction.update`.

        As a (quick) reminder: this dictionary has for keys:

          - "injection" (optional): a dictionary with keys (optional) "load_p", "load_q", "prod_p", "prod_v"
          - "hazards" (optional) : the outage suffered from the _grid
          - "maintenance" (optional) : the maintenance operations planned on the grid for the current time step.

        Returns
        -------
        timestamp: ``datetime.datetime``
            The current timestamp for which the modifications have been generated.

        dict_: ``dict``
            Always empty, indicating i do nothing (for this case)

        maintenance_time: ``numpy.ndarray``, dtype:``int``
            Information about the next planned maintenance. See :attr:`GridValue.maintenance_time` for more information.

        maintenance_duration: ``numpy.ndarray``, dtype:``int``
            Information about the duration of next planned maintenance. See :attr:`GridValue.maintenance_duration`
            for more information.

        hazard_duration: ``numpy.ndarray``, dtype:``int``
            Information about the current hazard. See :attr:`GridValue.hazard_duration`
            for more information.

        prod_v: ``numpy.ndarray``, dtype:``float``
            the (stored) value of the generator voltage setpoint

        Raises
        ------
        StopIteration
            if the chronics is over

        """
        self.current_datetime += self.time_interval
        return (
            self.current_datetime,
            {},
            self.maintenance_time,
            self.maintenance_duration,
            self.hazard_duration,
            None,
        )

    @abstractmethod
    def check_validity(self, backend):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

            This is called at the creation of the environment to ensure the Backend and the chronics
            are consistent with one another.

        A call to this method ensure that the action that will be sent to the current :class:`grid2op.Environment`
        can be properly implemented by its :class:`grid2op.Backend`.
        This specific method check that the dimension of all vectors are consistent

        Parameters
        ----------
        backend: :class:`grid2op.Backend.Backend`
            The backend used by the :class:`grid2op.Environment.Environment`
        """
        raise EnvError("check_validity not implemented")

    def done(self):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

            Use the :class:`ChroncisHandler` for such purpose

        Whether the episode is over or not.

        Returns
        -------
        done: ``bool``
            ``True`` means the episode has arrived to the end (no more data to generate) ``False`` means that the episode
            is not over yet.

        """
        if self.max_iter >= 0:
            return self.curr_iter >= self.max_iter
        else:
            return False

    def forecasts(self):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

            Use the :class:`ChroncisHandler` for such purpose

        This method is used to generate the forecasts that are made available to the :class:`grid2op.BaseAgent`.
        This forecasts are behaving the same way than a list of tuple as the one returned by
        :func:`GridValue.load_next` method.

        The way they are generated depends on the GridValue class. If not forecasts are made available, then
        the empty list should be returned.

        Returns
        -------
        res: ``list``
            Each element of this list having the same type as what is returned by :func:`GridValue.load_next`.

        """
        return []

    @abstractmethod
    def next_chronics(self):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

            Move to the next "chronics", representing the next "level" if we make the parallel
            with video games.

        A call to this function should at least restart:

          - :attr:`GridValue.current_datetime` to its origin value
          - :attr:`GridValue.curr_iter`

        """
        pass

    def tell_id(self, id_num, previous=False):
        """
        Tell the backend to use one folder for the chronics in particular. This method is mainly use when the GridValue
        object can deal with many folder. In this case, this method is used by the :class:`grid2op.Runner` to indicate
        which chronics to load for the current simulated episode.

        This is important to ensure reproducibility, especially in parrallel computation settings.

        This should also be used in case of generation "on the fly" of the chronics to ensure the same property.

        By default it does nothing.

        .. note::

            As of grid2op 1.6.4, this function now accepts the return value of `self.get_id()`.
        """
        warnings.warn(
            'Class {} doesn\'t handle different input folder. "tell_id" method has no impact.'
            "".format(type(self).__name__)
        )

    def get_id(self) -> str:
        """
        Utility to get the current name of the path of the data are looked at, if data are files.

        This could also be used to return a unique identifier to the generated chronics even in the case where they are
        generated on the fly, for example by return a hash of the seed.

        Returns
        -------
        res: ``str``
            A unique identifier of the chronics generated for this episode. For example, if the chronics comes from a
            specific folder, this could be the path to this folder.

        """
        warnings.warn(
            'Class {} doesn\'t handle different input folder. "get_id" method will return "".'
            "".format(type(self).__name__)
        )
        return ""

    def max_timestep(self):
        """
        This method returned the maximum timestep that the current episode can last.
        Note that if the :class:`grid2op.BaseAgent` performs a bad action that leads to a game over, then the episode
        can lasts less.

        Returns
        -------
        res: ``int``
            -1 if possibly infinite length or a positive integer representing the maximum duration of this episode

        """
        return self.max_iter

    def shuffle(self, shuffler=None):
        """
        This method can be overridden if the data that are represented by this object need to be shuffle.

        By default it does nothing.

        Parameters
        ----------
        shuffler: ``object``
            Any function that can be used to shuffle the data.

        """
        pass

    def sample_next_chronics(self, probabilities=None):
        """
        this is used to sample the next chronics used with given probabilities

        Parameters
        -----------
        probabilities: ``np.ndarray``
            Array of integer with the same size as the number of chronics in the cache.
            If it does not sum to one, it is rescaled such that it sums to one.

        Returns
        -------
        selected: ``int``
            The integer that was selected.

        Examples
        --------

        Let's assume in your chronics, the folder names are "Scenario_august_dummy", and
        "Scenario_february_dummy". For the sake of the example, we want the environment to loop
        75% of the time to the month of february and 25% of the time to the month of august.

        .. code-block:: python

            import grid2op
            env = grid2op.make("l2rpn_neurips_2020_track1", test=True)  # don't add "test=True" if
            # you don't want to perform a test.

            # check at which month will belong each observation
            for i in range(10):
                obs = env.reset()
                print(obs.month)
                # it always alternatively prints "8" (if chronics if from august) or
                # "2" if chronics is from february) with a probability of 50% / 50%

            env.seed(0)  # for reproducible experiment
            for i in range(10):
                _ = env.chronics_handler.sample_next_chronics([0.25, 0.75])
                obs = env.reset()
                print(obs.month)
                # it prints "2" with probability 0.75 and "8" with probability 0.25

        """

        return -1

    def set_filter(self, filter_fun):
        """
        Assign a filtering function to remove some chronics from the next time a call to "reset_cache" is called.

        **NB** filter_fun is applied to all element of :attr:`Multifolder.subpaths`. If ``True`` then it will
        be put in cache, if ``False`` this data will NOT be put in the cache.

        **NB** this has no effect until :attr:`Multifolder.reset` is called.

        Notes
        ------
        As of now, this has no effect unless the chronics are generated using :class:`Multifolder`
        or :class:`MultifolderWithCache`

        Examples
        --------
        Let's assume in your chronics, the folder names are "Scenario_august_dummy", and
        "Scenario_february_dummy". For the sake of the example, we want the environment to loop
        only through the month of february, because why not. Then we can do the following:

        .. code-block:: python

            import re
            import grid2op
            env = grid2op.make("l2rpn_neurips_2020_track1", test=True)  # don't add "test=True" if
            # you don't want to perform a test.

            # check at which month will belong each observation
            for i in range(10):
                obs = env.reset()
                print(obs.month)
                # it always alternatively prints "8" (if chronics if from august) or
                # "2" if chronics is from february)

            # to see where the chronics are located
            print(env.chronics_handler.subpaths)

            # keep only the month of february
            env.chronics_handler.set_filter(lambda path: re.match(".*february.*", path) is not None)
            env.chronics_handler.reset()  # if you don't do that it will not have any effect

            for i in range(10):
                obs = env.reset()
                print(obs.month)
                # it always prints "2" (representing february)

        """
        warnings.warn(
            f'Calling this function has no effect for chronics generated from "{type(self)}"'
        )

    def set_chunk_size(self, new_chunk_size):
        """
        This parameters allows to set, if the data generation process support it, the amount of data that is read
        at the same time. It can help speeding up the computation process by adding more control on the io operation.

        Parameters
        ----------
        new_chunk_size: ``int``
            The chunk size (ie the number of rows that will be read on each data set at the same time)

        """
        pass

    def fast_forward(self, nb_timestep):
        """
        INTERNAL

        .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

            Prefer using :func:`grid2op.Environment.BaseEnv.fast_forward_chronics`

        This method allows you to skip some time step at the beginning of the chronics.

        This is useful at the beginning of the training, if you want your agent to learn on more diverse scenarios.
        Indeed, the data provided in the chronics usually starts always at the same date time (for example Jan 1st at
        00:00). This can lead to suboptimal exploration, as during this phase, only a few time steps are managed by
        the agent, so in general these few time steps will correspond to grid state around Jan 1st at 00:00.

        Parameters
        ----------
        nb_timestep: ``int``
            Number of time step to "fast forward"

        """
        for _ in range(nb_timestep):
            self.load_next()



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import copy
import time

import numpy as np

from grid2op.Environment import Environment
from grid2op.Agent import BaseAgent

from grid2op.Episode import EpisodeData
from grid2op.Runner.FakePBar import _FakePbar
from grid2op.dtypes import dt_int, dt_float, dt_bool
from grid2op.Chronics import ChronicsHandler


def _aux_add_data(reward, env, episode,
                  efficient_storing, end__, beg__, act,
                  obs, info, time_step, opp_attack):
    episode.incr_store(
        efficient_storing,
        time_step,
        end__ - beg__,
        float(reward),
        env._env_modification,
        act,
        obs,
        opp_attack,
        info,
    )
    return reward
                
def _aux_one_process_parrallel(
    runner,
    episode_this_process,
    process_id,
    path_save=None,
    env_seeds=None,
    agent_seeds=None,
    max_iter=None,
    add_detailed_output=False,
    add_nb_highres_sim=False,
):
    """this is out of the runner, otherwise it does not work on windows / macos"""
    chronics_handler = ChronicsHandler(
        chronicsClass=runner.gridStateclass,
        path=runner.path_chron,
        **runner.gridStateclass_kwargs
    )
    parameters = copy.deepcopy(runner.parameters)
    nb_episode_this_process = len(episode_this_process)
    res = [(None, None, None) for _ in range(nb_episode_this_process)]
    for i, ep_id in enumerate(episode_this_process):
        # `ep_id`: grid2op id of the episode i want to play
        # `i`: my id of the episode played (0, 1, ... episode_this_process)
        env, agent = runner._new_env(
            chronics_handler=chronics_handler, parameters=parameters
        )
        try:
            env_seed = None
            if env_seeds is not None:
                env_seed = env_seeds[i]
            agt_seed = None
            if agent_seeds is not None:
                agt_seed = agent_seeds[i]
            tmp_ = _aux_run_one_episode(
                env,
                agent,
                runner.logger,
                ep_id,
                path_save,
                env_seed=env_seed,
                max_iter=max_iter,
                agent_seed=agt_seed,
                detailed_output=add_detailed_output,
            )
            (name_chron, cum_reward, nb_time_step, max_ts, episode_data, nb_highres_sim)  = tmp_
            id_chron = chronics_handler.get_id()
            res[i] = (id_chron, name_chron, float(cum_reward), nb_time_step, max_ts)
            
            if add_detailed_output:
                res[i] = (*res[i], episode_data)
            if add_nb_highres_sim:
                res[i] = (*res[i], nb_highres_sim)                
        finally:
            env.close()
    return res


def _aux_run_one_episode(
    env: Environment,
    agent: BaseAgent,
    logger,
    indx : int,
    path_save=None,
    pbar=False,
    env_seed=None,
    agent_seed=None,
    max_iter=None,
    detailed_output=False,
):
    done = False
    time_step = int(0)
    time_act = 0.0
    cum_reward = dt_float(0.0)

    # set the environment to use the proper chronic
    env.set_id(indx)
    # set the seed
    if env_seed is not None:
        env.seed(env_seed)

    # handle max_iter
    if max_iter is not None:
        env.chronics_handler.set_max_iter(max_iter)

    # reset it
    obs = env.reset()
    # reset the number of calls to high resolution simulator
    env._highres_sim_counter._HighResSimCounter__nb_highres_called = 0
    
    # seed and reset the agent
    if agent_seed is not None:
        agent.seed(agent_seed)
    agent.reset(obs)

    # compute the size and everything if it needs to be stored
    nb_timestep_max = env.chronics_handler.max_timestep()
    efficient_storing = nb_timestep_max > 0
    nb_timestep_max = max(nb_timestep_max, 0)
    max_ts = nb_timestep_max
    if path_save is None and not detailed_output:
        # i don't store anything on drive, so i don't need to store anything on memory
        nb_timestep_max = 0

    disc_lines_templ = np.full((1, env.backend.n_line), fill_value=False, dtype=dt_bool)

    attack_templ = np.full(
        (1, env._oppSpace.action_space.size()), fill_value=0.0, dtype=dt_float
    )
    
    if efficient_storing:
        times = np.full(nb_timestep_max, fill_value=np.NaN, dtype=dt_float)
        rewards = np.full(nb_timestep_max, fill_value=np.NaN, dtype=dt_float)
        actions = np.full(
            (nb_timestep_max, env.action_space.n), fill_value=np.NaN, dtype=dt_float
        )
        env_actions = np.full(
            (nb_timestep_max, env._helper_action_env.n),
            fill_value=np.NaN,
            dtype=dt_float,
        )
        observations = np.full(
            (nb_timestep_max + 1, env.observation_space.n),
            fill_value=np.NaN,
            dtype=dt_float,
        )
        disc_lines = np.full(
            (nb_timestep_max, env.backend.n_line), fill_value=np.NaN, dtype=dt_bool
        )
        attack = np.full(
            (nb_timestep_max, env._opponent_action_space.n),
            fill_value=0.0,
            dtype=dt_float,
        )
        legal = np.full(nb_timestep_max, fill_value=True, dtype=dt_bool)
        ambiguous = np.full(nb_timestep_max, fill_value=False, dtype=dt_bool)
    else:
        times = np.full(0, fill_value=np.NaN, dtype=dt_float)
        rewards = np.full(0, fill_value=np.NaN, dtype=dt_float)
        actions = np.full((0, env.action_space.n), fill_value=np.NaN, dtype=dt_float)
        env_actions = np.full(
            (0, env._helper_action_env.n), fill_value=np.NaN, dtype=dt_float
        )
        observations = np.full(
            (0, env.observation_space.n), fill_value=np.NaN, dtype=dt_float
        )
        disc_lines = np.full((0, env.backend.n_line), fill_value=np.NaN, dtype=dt_bool)
        attack = np.full(
            (0, env._opponent_action_space.n), fill_value=0.0, dtype=dt_float
        )
        legal = np.full(0, fill_value=True, dtype=dt_bool)
        ambiguous = np.full(0, fill_value=False, dtype=dt_bool)

    need_store_first_act = path_save is not None or detailed_output
    if need_store_first_act:
        # store observation at timestep 0
        if efficient_storing:
            observations[time_step, :] = obs.to_vect()
        else:
            observations = np.concatenate((observations, obs.to_vect().reshape(1, -1)))
            
    episode = EpisodeData(
        actions=actions,
        env_actions=env_actions,
        observations=observations,
        rewards=rewards,
        disc_lines=disc_lines,
        times=times,
        observation_space=env.observation_space,
        action_space=env.action_space,
        helper_action_env=env._helper_action_env,
        path_save=path_save,
        disc_lines_templ=disc_lines_templ,
        attack_templ=attack_templ,
        attack=attack,
        attack_space=env._opponent_action_space,
        logger=logger,
        name=env.chronics_handler.get_name(),
        force_detail=detailed_output,
        other_rewards=[],
        legal=legal,
        ambiguous=ambiguous,
        has_legal_ambiguous=True,
    )
    if need_store_first_act:
        # I need to manually force in the first observation (otherwise it's not computed)
        episode.observations.objects[0] = episode.observations.helper.from_vect(
            observations[time_step, :]
        )
    episode.set_parameters(env)

    beg_ = time.perf_counter()

    reward = float(env.reward_range[0])
    done = False

    next_pbar = [False]
    with _aux_make_progress_bar(pbar, nb_timestep_max, next_pbar) as pbar_:
        while not done:
            beg__ = time.perf_counter()
            act = agent.act(obs, reward, done)
            end__ = time.perf_counter()
            time_act += end__ - beg__
            
            if type(env).CAN_SKIP_TS:
                # the environment can "skip" some time
                # steps I need to call the 'env.steps()' to get all
                # the steps.
                res_env_tmp = env.steps(act)
                for (obs, reward, done, info), opp_attack in zip(*res_env_tmp):
                    time_step += 1
                    cum_reward += _aux_add_data(reward, env, episode,
                                                efficient_storing,
                                                end__, beg__, act,
                                                obs, info, time_step,
                                                opp_attack)
                    pbar_.update(1)
            else:
                # regular environment
                obs, reward, done, info = env.step(act)
                time_step += 1
                opp_attack = env._oppSpace.last_attack
                cum_reward += _aux_add_data(reward, env, episode,
                                            efficient_storing,
                                            end__, beg__, act,
                                            obs, info, time_step,
                                            opp_attack)
                pbar_.update(1)

        end_ = time.perf_counter()
    episode.set_meta(env, time_step, float(cum_reward), env_seed, agent_seed)
    li_text = [
        "Env: {:.2f}s",
        "\t - apply act {:.2f}s",
        "\t - run pf: {:.2f}s",
        "\t - env update + observation: {:.2f}s",
        "Agent: {:.2f}s",
        "Total time: {:.2f}s",
        "Cumulative reward: {:1f}",
    ]
    msg_ = "\n".join(li_text)
    logger.info(
        msg_.format(
            env._time_apply_act + env._time_powerflow + env._time_extract_obs,
            env._time_apply_act,
            env._time_powerflow,
            env._time_extract_obs,
            time_act,
            end_ - beg_,
            cum_reward,
        )
    )

    episode.set_episode_times(env, time_act, beg_, end_)

    episode.to_disk()
    name_chron = env.chronics_handler.get_name()
    return (name_chron, cum_reward,
            int(time_step),
            int(max_ts),
            episode,
            env.nb_highres_called)


def _aux_make_progress_bar(pbar, total, next_pbar):
    """
    INTERNAL

    .. warning:: /!\\\\ Internal, do not use unless you know what you are doing /!\\\\

    Parameters
    ----------
    pbar: ``bool`` or ``type`` or ``object``
        How to display the progress bar, understood as follow:

        - if pbar is ``None`` nothing is done.
        - if pbar is a boolean, tqdm pbar are used, if tqdm package is available and installed on the system
          [if ``true``]. If it's false it's equivalent to pbar being ``None``
        - if pbar is a ``type`` ( a class), it is used to build a progress bar at the highest level (episode) and
          and the lower levels (step during the episode). If it's a type it muyst accept the argument "total"
          and "desc" when being built, and the closing is ensured by this method.
        - if pbar is an object (an instance of a class) it is used to make a progress bar at this highest level
          (episode) but not at lower levels (step during the episode)
    """
    pbar_ = _FakePbar()
    next_pbar[0] = False

    if isinstance(pbar, bool):
        if pbar:
            try:
                from tqdm import tqdm

                pbar_ = tqdm(total=total, desc="episode")
                next_pbar[0] = True
            except (ImportError, ModuleNotFoundError):
                pass
    elif isinstance(pbar, type):
        pbar_ = pbar(total=total, desc="episode")
        next_pbar[0] = pbar
    elif isinstance(pbar, object):
        pbar_ = pbar
    return pbar_



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np

from grid2op.dtypes import dt_bool
from grid2op.Agent.greedyAgent import GreedyAgent


class PowerLineSwitch(GreedyAgent):
    """
    This is a :class:`GreedyAgent` example, which will attempt to disconnect powerlines.

    It will choose among:

      - doing nothing
      - changing the status of one powerline

    which action that will maximize the simulated reward. All powerlines are tested at each steps. This means
    that if `n` is the number of powerline on the grid, at each steps this actions will perform `n` +1
    calls to "simulate" (one to do nothing and one that change the status of each powerline)

    """

    def __init__(self, action_space):
        GreedyAgent.__init__(self, action_space)

    def _get_tested_action(self, observation):
        res = [self.action_space({})]  # add the do nothing
        for i in range(self.action_space.n_line):
            tmp = np.full(self.action_space.n_line, fill_value=False, dtype=dt_bool)
            tmp[i] = True
            action = self.action_space({"change_line_status": tmp})
            if not observation.line_status[i]:
                # so the action consisted in reconnecting the powerline
                # i need to say on which bus (always on bus 1 for this type of agent)
                action = action.update(
                    {"set_bus": {"lines_or_id": [(i, 1)], "lines_ex_id": [(i, 1)]}}
                )
            res.append(action)
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Agent.baseAgent import BaseAgent


class OneChangeThenNothing(BaseAgent):
    """
    This is a specific kind of BaseAgent. It does an BaseAction (possibly non empty) at the first time step and then does
    nothing.

    This class is an abstract class and cannot be instanciated (ie no object of this class can be created). It must
    be overridden and the method :func:`OneChangeThenNothing._get_dict_act` be defined. Basically, it must know
    what action to do.

    Attributes
    ------------
    my_dict: ``dict`` (class member)
        Representation, as a dictionnary of the only action that this Agent will do at the first time step.

    Examples
    ---------
    We advise to use this class as following

    .. code-block:: python

        import grid2op
        from grid2op.Agent import OneChangeThenNothing
        acts_dict_ = [{}, {"set_line_status": [(0,-1)]}]  # list of dictionaries. Each dictionary
        # represents a valid action

        env = grid2op.make()  # create an environment
        for act_as_dict in zip(acts_dict_):
            # generate the proper class that will perform the first action (encoded by {}) in acts_dict_
            agent_class = OneChangeThenNothing.gen_next(act_as_dict)

            # start a runner with this agent
            runner = Runner(**env.get_params_for_runner(), agentClass=agent_class)
            # run 2 episode with it
            res_2 = runner.run(nb_episode=2)

    """

    my_dict = {}

    def __init__(self, action_space):
        BaseAgent.__init__(self, action_space)
        self.has_changed = False
        self.do_nothing_action = self.action_space({})

    def act(self, observation, reward, done=False):
        if self.has_changed:
            res = self.do_nothing_action
        else:
            res = self.action_space(self._get_dict_act())
            self.has_changed = True
        return res

    def reset(self, obs):
        self.has_changed = False

    def _get_dict_act(self):
        """
        Function that need to be overridden to indicate which action to perform.

        Returns
        -------
        res: ``dict``
            A dictionnary that can be converted into a valid :class:`grid2op.BaseAction.BaseAction`. See the help of
            :func:`grid2op.BaseAction.ActionSpace.__call__` for more information.
        """
        return self.my_dict

    @classmethod
    def gen_next(cls, dict_):
        """
        This function allows to change the dictionnary of the action that the agent will perform.

        See the class level documentation for an example on how to use this.

        Parameters
        ----------
        dict_: ``dict``
            A dictionnary representing an action. This dictionnary is assumed to be convertible into an action.
            No check is performed at this stage.


        """
        cls.my_dict = dict_
        return cls



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from abc import abstractmethod
import numpy as np
from grid2op.Agent.baseAgent import BaseAgent
from grid2op.dtypes import dt_float


class GreedyAgent(BaseAgent):
    """
    This is a class of "Greedy BaseAgent". Greedy agents are all executing the same kind of algorithm to take action:

      1. They :func:`grid2op.Observation.Observation.simulate` all actions in a given set
      2. They take the action that maximise the simulated reward among all these actions

    This class is an abstract class (object of this class cannot be created). To create "GreedyAgent" one must
    override this class. Examples are provided with :class:`PowerLineSwitch` and :class:`TopologyGreedy`.
    """

    def __init__(self, action_space):
        BaseAgent.__init__(self, action_space)
        self.tested_action = None
        self.resulting_rewards = None

    def act(self, observation, reward, done=False):
        """
        By definition, all "greedy" agents are acting the same way. The only thing that can differentiate multiple
        agents is the actions that are tested.

        These actions are defined in the method :func:`._get_tested_action`. This :func:`.act` method implements the
        greedy logic: take the actions that maximizes the instantaneous reward on the simulated action.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The current observation of the :class:`grid2op.Environment.Environment`

        reward: ``float``
            The current reward. This is the reward obtained by the previous action

        done: ``bool``
            Whether the episode has ended or not. Used to maintain gym compatibility

        Returns
        -------
        res: :class:`grid2op.Action.Action`
            The action chosen by the bot / controller / agent.

        """
        self.tested_action = self._get_tested_action(observation)
        if len(self.tested_action) > 1:
            self.resulting_rewards = np.full(
                shape=len(self.tested_action), fill_value=np.NaN, dtype=dt_float
            )
            for i, action in enumerate(self.tested_action):
                (
                    simul_obs,
                    simul_reward,
                    simul_has_error,
                    simul_info,
                ) = observation.simulate(action)
                self.resulting_rewards[i] = simul_reward
            reward_idx = int(
                np.argmax(self.resulting_rewards)
            )  # rewards.index(max(rewards))
            best_action = self.tested_action[reward_idx]
        else:
            best_action = self.tested_action[0]
        return best_action

    @abstractmethod
    def _get_tested_action(self, observation):
        """
        Returns the list of all the candidate actions.

        From this list, the one that achieve the best "simulated reward" is used.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The current observation of the :class:`grid2op.Environment.Environment`

        Returns
        -------
        res: ``list``
            A list of all candidate :class:`grid2op.BaseAction.BaseAction`
        """
        pass



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

import numpy as np

from grid2op.Exceptions import Grid2OpException
from grid2op.Reward.baseReward import BaseReward
from grid2op.dtypes import dt_float


class AlarmReward(BaseReward):
    """
    This reward is based on the "alarm feature" where the agent is asked to send information about potential issue
    on the grid.

    On this case, when the environment is in a "game over" state (eg it's the end) then the reward is computed
    the following way:

    - if the environment has been successfully manage until the end of the chronics, then 1.0 is returned
    - if no alarm has been raised, then -1.0 is return


    Examples
    ---------
    You can use this reward in any environment with:

    .. code-block:

        import grid2op
        from grid2op.Reward import AlarmReward

        # then you create your environment with it:
        NAME_OF_THE_ENVIRONMENT = "rte_case14_realistic"
        env = grid2op.make(NAME_OF_THE_ENVIRONMENT,reward_class=AlarmReward)
        # and do a step with a "do nothing" action
        obs = env.reset()
        obs, reward, done, info = env.step(env.action_space())
        # the reward is computed with the AlarmReward class


    """

    def __init__(self, logger=None):
        BaseReward.__init__(self, logger=logger)
        # required if you want to design a custom reward taking into account the
        # alarm feature
        self.has_alarm_component = True
        self.is_alarm_used = False  # required to update it in __call__ !!

        self.total_time_steps = dt_float(0.0)
        self.reward_min = dt_float(-1.0)
        self.reward_max = dt_float(1.0)
        self.reward_no_game_over = dt_float(0.0)

        self.window_size = None
        self.best_time = None

        self.mult_for_right_zone = 2

    def initialize(self, env):
        if not env._has_attention_budget:
            raise Grid2OpException(
                'Impossible to use the "AlarmReward" with an environment for which this feature '
                'is disabled. Please make sure "env._has_attention_budget" is set to ``True`` or '
                "change the reward class with `grid2op.make(..., reward_class=AnyOtherReward)`"
            )
        self.reset(env)

    def reset(self, env):
        self.total_time_steps = env.max_episode_duration()
        self.best_time = env.parameters.ALARM_BEST_TIME
        self.window_size = env.parameters.ALARM_WINDOW_SIZE

    def _tmp_score_time(self, step_alarm, step_game_over):
        """
        compute the "temporal" score.

        Should give a number between 0 and 1
        """
        if step_game_over - step_alarm > self.best_time + self.window_size:
            # alarm too soon
            res = 0
        elif step_game_over - step_alarm < self.best_time - self.window_size:
            # alarm too late
            res = 0
        else:
            # square function such that: it gives 1 if step_game_over - step_alarm equals self.best_time
            # and 0 if  step_game_over - step_alarm = self.best_time + self.window_size or
            # if step_game_over - step_alarm self.best_time - self.window_size
            dist_to_game_over = step_game_over - step_alarm
            dist_to_best = dist_to_game_over - self.best_time

            # set it to 0 for the extreme case
            polynom = (dist_to_best - self.window_size) * (
                dist_to_best + self.window_size
            )
            # scale it such that it is 1 for dist_to_best == 0 (ie step_game_over - step_alarm == self.best_time)
            res = -polynom / self.window_size**2
        return res

    def _mult_for_zone(self, alarm, disc_lines, env):
        """compute the multiplicative factor that increases the score if the right zone is predicted"""
        res = 1.0
        # extract the lines that have been disconnected due to cascading failures
        lines_disconnected_first = np.where(disc_lines == 0)[0]

        if (
            alarm.sum() > 1
        ):  # if we have more than one zone in the alarm, we cannot discrtiminate, no bonus points
            return res

        # extract the zones they belong too
        zones_these_lines = set()
        zone_for_each_lines = env.alarms_lines_area
        for line_id in lines_disconnected_first:
            line_name = env.name_line[line_id]
            for zone_name in zone_for_each_lines[line_name]:
                zones_these_lines.add(zone_name)

        # now retrieve the id of the zones in which a powerline has been disconnected
        list_zone_names = list(zones_these_lines)
        list_zone_ids = np.where(np.isin(env.alarms_area_names, list_zone_names))[0]
        # and finally, award some extra points if one of the zone, containing one of the powerline disconnected
        # by protection is in the alarm
        if alarm[list_zone_ids].any():
            res *= self.mult_for_right_zone
        return res

    def _points_for_alarm(self, step_alarm, alarm, step_game_over, disc_lines, env):
        """how much points are given for this specific alarm"""
        is_alarm_used = False
        score = self.reward_min
        score_for_time = self._tmp_score_time(step_alarm, step_game_over)
        if score_for_time != 0:
            is_alarm_used = True  # alarm is in the right time window
            score = score_for_time
            score *= (
                self._mult_for_zone(alarm, disc_lines, env) / self.mult_for_right_zone
            )
        return score, is_alarm_used

    def __call__(self, action, env, has_error, is_done, is_illegal, is_ambiguous):
        if is_done:
            if not has_error:
                # agent went until the end
                return self.reward_max

            if np.all(env._disc_lines == -1):
                # game over is not caused by the tripping of a powerline
                return self.reward_min

            if len(env._attention_budget._all_successful_alarms) == 0:
                # no alarm have been sent, so it's the minimum
                return self.reward_min

            successfull_alarms = env._attention_budget._all_successful_alarms
            step_game_over = env.nb_time_step
            disc_lines = env._disc_lines

            # so now i can consider the alarms.
            best_score, is_alarm_used = self.reward_min, False
            for alarm in successfull_alarms:
                tmp_sc, tmp_is = self._points_for_alarm(
                    *alarm,
                    step_game_over=step_game_over,
                    disc_lines=disc_lines,
                    env=env
                )
                if tmp_sc > best_score:
                    best_score = tmp_sc
                    is_alarm_used = tmp_is

            self.is_alarm_used = is_alarm_used
            return best_score
        else:
            res = self.reward_no_game_over
        return res



# Copyright (c) 2019-2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

from grid2op.Agent.baseAgent import BaseAgent


class DoNothingAgent(BaseAgent):
    """
    This is the most basic BaseAgent. It is purely passive, and does absolutely nothing.

    As opposed to most reinforcement learning environments, in grid2op, doing nothing is often
    the best solution.

    """

    def __init__(self, action_space):
        BaseAgent.__init__(self, action_space)

    def act(self, observation, reward, done=False):
        """
        As better explained in the document of :func:`grid2op.BaseAction.update` or
        :func:`grid2op.BaseAction.ActionSpace.__call__`.

        The preferred way to make an object of type action is to call :func:`grid2op.BaseAction.ActionSpace.__call__`
        with the dictionary representing the action. In this case, the action is "do nothing" and it is represented by
        the empty dictionary.

        Parameters
        ----------
        observation: :class:`grid2op.Observation.Observation`
            The current observation of the :class:`grid2op.Environment.Environment`

        reward: ``float``
            The current reward. This is the reward obtained by the previous action

        done: ``bool``
            Whether the episode has ended or not. Used to maintain gym compatibility

        Returns
        -------
        res: :class:`grid2op.Action.Action`
            The action chosen by the bot / controller / agent.

        """
        res = self.action_space({})
        return res



# Copyright (c) 2019-2022, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of Grid2Op, Grid2Op a testbed platform to model sequential decision making in power systems.

__all__ = [
    "ActionWCCI2020", "ObservationWCCI2020",
    "ActionNeurips2020", "ObservationNeurips2020",
    "ActionICAPS2021", "ObservationICAPS2021",
    "ActionWCCI2022", "ObservationWCCI2022",
    "ActionIDF2023", "ObservationIDF2023"
    ]

from grid2op.l2rpn_utils.wcci_2020 import ActionWCCI2020, ObservationWCCI2020
from grid2op.l2rpn_utils.neurips_2020 import ActionNeurips2020, ObservationNeurips2020
from grid2op.l2rpn_utils.icaps_2021 import ActionICAPS2021, ObservationICAPS2021
from grid2op.l2rpn_utils.wcci_2022 import ActionWCCI2022, ObservationWCCI2022
from grid2op.l2rpn_utils.idf_2023 import ActionIDF2023, ObservationIDF2023